// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef WORLDMODELHIGHLEVEL_H
#define WORLDMODELHIGHLEVEL_H

#include <rcsc/geom/triangle_2d.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/circle_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/line_2d.h>
#include <rcsc/geom/segment_2d.h>
#include <rcsc/player/player_object.h>
#include <rcsc/player/player_agent.h>
#include <iostream>
#include <math.h>
#include <rcsc/math_util.h>
#include <rcsc/soccer_math.h>
#include <rcsc/common/server_param.h>


#define PITCH_WIDTH  68.0
#define PITCH_LENGTH 105.0

using namespace rcsc;
using namespace std ;

/**
	@author <Hosein@Mobasher>
	all right reserved by robosampad team.
	Copy Right (c) RoboSampad 2D Soccer Simulation Team.
*/

#define EPSILON 0.0001  /*!< Value used for floating point equality tests. */
typedef double AngRad;  /*!< Type definition for angles in degrees. */


enum setType{
	opponent , teammate
};

enum HandSide
{
	HS_RIGHT , HS_LEFT , HS_ILLEGAL
};

class WorldModelHighLevel{

private :
	int nr , counter ;
	double dist;
	bool   isIn;

public:

WorldModelHighLevel(): nr(0) , counter(0) , dist(0.0) , isIn(false)
{
}


~WorldModelHighLevel()
{
}

/*-------------------------------------------------------------------*/
/*!  \brief return teammate type by unum */
const
PlayerObject * getTeammate( const int unum , PlayerAgent * agent ) const
{
	const rcsc::WorldModel & wm = agent->world();// world model object

	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it =  wm.teammatesFromSelf().begin();
		it != end;
		++it )
	{
	if ( (*it)->unum() == unum )
	{
		return *it;
	}
	}
	return static_cast< PlayerObject * >( 0 );
}


PlayerObject * getTeammate( const int unum , PlayerAgent * agent )
{
	const rcsc::WorldModel & wm = agent->world();// world model object

	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it =  wm.teammatesFromSelf().begin();
		it != end;
		++it )
	{
	if ( (*it)->unum() == unum )
	{
		return *it;
	}
	}
	return static_cast< PlayerObject * >( 0 );
}

/*-------------------------------------------------------------------*/
/*!  \brief return opponent type by unum */
const
PlayerObject * getOpponent( const int unum , PlayerAgent * agent ) const
{

	const rcsc::WorldModel & wm = agent->world();// world model object

	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it =  wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
	if ( (*it)->unum() == unum )
	{
		return *it;
	}
	}
	return static_cast< PlayerObject * >( 0 );
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns the number of visible objects that
     \brief located in the triangle 't'.
.*/

int getNrInSetInTriangle( PlayerAgent * agent , Triangle2D t , bool tmm = false )
{
	nr = 0 ;
	double dist_thr = agent->world().self().pos().dist(t.centroid())*0.1;
	if( dist_thr > 1.0 ) dist_thr = 1.0;

	if( tmm )
		nr =  agent->world().countTeammatesIn( t , int(dist_thr) , true );
	else
		nr =  agent->world().countOpponentsIn( t , int(dist_thr) , true );
// 	const rcsc::WorldModel & agent->world() = agent->world();// world model object
/*
	if( !tmm )
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getOpponent( counter + 1 , agent ) )
			{
				players    [counter] = getOpponent( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && t.contains(players_pos[counter]) )
				nr ++;
	}
	else
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getTeammate( counter + 1 , agent ) )
			{
				players    [counter] = getTeammate( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && t.contains(players_pos[counter]) )
				nr ++;
	}
*/
	return nr;
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns the number of visible objects that
     \brief located in the circle 'c'.
.*/

int getNrInSetInCircle( PlayerAgent * agent ,  Circle2D   c , bool tmm = false )
{
	nr = 0 ;

	double dist_thr = agent->world().self().pos().dist(c.center())*0.1;
	if( dist_thr > 1.0 ) dist_thr = 1.0;

	if( tmm )
		nr =  agent->world().countTeammatesIn( c , int(dist_thr) , true );
	else
		nr =  agent->world().countOpponentsIn( c , int(dist_thr) , true );
// 	const rcsc::WorldModel & agent->world() = agent->world();// world model object
/*
	if( !tmm )
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getOpponent( counter + 1 , agent ) )
			{
				players    [counter] = getOpponent( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && c.contains(players_pos[counter]) )
				nr ++;
	}
	else
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getTeammate( counter + 1 , agent ) )
			{
				players    [counter] = getTeammate( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && c.contains(players_pos[counter]) )
				nr ++;
	}*/
	return nr;
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns the number of visible objects that
     \brief located in the rect 'r'.
.*/


int getNrInSetInRect( PlayerAgent * agent , Rect2D r , bool tmm = false )
{
	nr = 0 ;
	double dist_thr = agent->world().self().pos().dist(r.center())*0.1;
	if( dist_thr > 1.0 ) dist_thr = 1.0;

	if( tmm )
		nr =  agent->world().countTeammatesIn( r , int(dist_thr) , true );
	else
		nr =  agent->world().countOpponentsIn( r , int(dist_thr) , true );

// 	const rcsc::WorldModel & agent->world() = agent->world();// world model object
/*
	if( !tmm )
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getOpponent( counter + 1 , agent ) )
			{
				players    [counter] = getOpponent( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && r.contains(players_pos[counter]) )
				nr ++;
	}
	else
	{
		const PlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( getTeammate( counter + 1 , agent ) )
			{
				players    [counter] = getTeammate( counter + 1 , agent );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && r.contains(players_pos[counter]) )
				nr ++;
	}*/

	return nr;
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns the vector position is in triangle 't' or not
.*/

bool isVector2DInTriangle( Triangle2D t , Vector2D v )
{
	return t.contains(v);
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns line 'l' distance to vector position
.*/

double getLineDistanceToVector( Line2D     l , Vector2D v )
{
	dist = 0.0;
	dist = l.dist(v);
	return dist;
}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns line 'l' and Circle 'c' intersection point
.*/

Vector2D getPosCircleAndLine( PlayerAgent * agent , Circle2D c , Line2D l )
{
	Vector2D posReturn;
	Vector2D posTemp;
	Vector2D posAgent = getAgentGlobalPosition(agent);
	c.intersection( l , &posReturn , &posTemp );

	return posReturn;

}

/*-------------------------------------------------------------------*/
/*!
     \brief This method returns agent global position
.*/

Vector2D getAgentGlobalPosition( PlayerAgent * agent )
{
	return agent->world().self().pos();
}

double getAgentStamina( PlayerAgent * agent )
{
	return agent->world().self().stamina();
}

bool isBallKickable ( PlayerAgent * agent )
{
	bool kickable = agent->world().self().isKickable();
	
	if ( agent->world().existKickableTeammate()
		&& agent->world().teammatesFromBall().front()->distFromBall()
		< agent->world().ball().distFromSelf() )
	{
		kickable = false;
	}
	
	return kickable;
}

bool isBallCatchable( PlayerAgent * agent )
{
	rcsc::GameTime M_last_goalie_kick_time;
	static const
		rcsc::Rect2D our_penalty( Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
					-rcsc::ServerParam::i().penaltyAreaHalfWidth() + 1.0 ),
					Vector2D( rcsc::ServerParam::i().penaltyAreaLength() - 1.0,
					rcsc::ServerParam::i().penaltyAreaWidth() - 2.0 ));
	
	if ( agent->world().time().cycle() > M_last_goalie_kick_time.cycle() + 5
		&& agent->world().ball().distFromSelf() < agent->world().self().catchableArea() - 0.05
		&& our_penalty.contains( agent->world().ball().pos() ) )
		return true;
	
		return false;
}

bool isDeadBallUs( PlayerAgent * agent )
{

	rcsc::GameMode::Type gm = agent->world().gameMode().type();
	const rcsc::WorldModel & world = agent->world();

	if( world.ourSide()  != world.gameMode().side())
		return false;
	else if(gm == rcsc::GameMode::IndFreeKick_  ||
		gm == rcsc::GameMode::CatchFault_   ||
		gm == rcsc::GameMode::FreeKickFault_||
		gm == rcsc::GameMode::BackPass_     ||
		gm == rcsc::GameMode::FoulBallOut_  ||
		gm == rcsc::GameMode::PenaltyKick_  ||
		gm == rcsc::GameMode::OffSide_      ||
		gm == rcsc::GameMode::CornerKick_   ||
		gm == rcsc::GameMode::FreeKick_     ||
		gm == rcsc::GameMode::KickIn_       ||
		gm == rcsc::GameMode::KickOff_      ||
		gm == rcsc::GameMode::CornerKick_   ||
		gm == rcsc::GameMode::BeforeKickOff   )
		return true;
	else
		return false;

}

bool isBeforeKickOffUs( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::BeforeKickOff &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isCornerKickUs   ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::CornerKick_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isKickInUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::KickIn_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isFreeKickUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::FreeKick_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isKickOffUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::KickOff_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isOffSideUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::OffSide_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isPenaltyKickUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::PenaltyKick_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isInFiled ( PlayerAgent * agent , Vector2D v )
{
	Rect2D rect( Vector2D( -52.5 , -34.0 ) , Vector2D( 52.5 , 34.0 ) );
	if( rect.contains( v ) )
		return true;
	else
		return false;
}

bool isBackPassUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::BackPass_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isIndirectFreeKickUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::IndFreeKick_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isCatchFaultUs ( PlayerAgent * agent )
{
	if( agent->world().gameMode().type() == rcsc::GameMode::CatchFault_ &&
	    agent->world().ourSide()  != agent->world().gameMode().side())
		return true;
	else
		return false;
}

bool isInOurPenaltyArea( PlayerAgent * agent , Vector2D pos )
{
	static const
	rcsc::Rect2D our_penalty( Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
				-rcsc::ServerParam::i().penaltyAreaHalfWidth() + 1.0),
				Vector2D(rcsc::ServerParam::i().penaltyAreaLength() - 1.0,
				rcsc::ServerParam::i().penaltyAreaWidth() - 2.0 ));

	return our_penalty.contains(pos);
}

bool isInTheirPenaltyArea( PlayerAgent * agent , Vector2D pos )
{
// 	static const
// 	rcsc::Rect2D their_penalty(  rcsc::ServerParam::i().pitchHalfLength(),
// 				     rcsc::ServerParam::i().penaltyAreaHalfWidth() + 1.0,
// 				     rcsc::ServerParam::i().penaltyAreaLength() - 1.0,
// 				     rcsc::ServerParam::i().penaltyAreaWidth() - 2.0 );
	rcsc::Rect2D their_penalty(  Vector2D( 52.5,  26.25),
				     Vector2D( 40.0, -26.25 )
			          );
	

	return their_penalty.contains(pos);
}

const PlayerObject * convertUnumToPlayerObject( rcsc::PlayerAgent * agent , int unum ) const 
{

     const rcsc::WorldModel & wm = agent->world();
     const PlayerObject * p ;

     const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();

     for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
           it != end;
           ++it )
     {
         if ( (*it)->unum() == 1 )
         {
             p = *it;
             break;
         }
     }

	return p;
// cout<<' '<<nearest_tmm->unum()<<endl;

}

bool isOnside( PlayerAgent * agent , int i  )
{
	const PlayerObject * p = getTeammate( i , agent );
	Vector2D       v( -1000 , 0 );

	if( p ) v = p->pos();

	if( agent->world().offsideLineX() > v.x )
		return true;
	else
		return false;
}

bool isOnside( PlayerAgent * agent , Vector2D v  )
{

	if( agent->world().offsideLineX() > v.x )
		return true;
	else
		return false;
}

bool isOnside( PlayerAgent * agent , const AbstractPlayerObject * player  )
{

	if( player )
	{
		if( agent->world().offsideLineX() > player->pos().x )
			return true;
		else
			return false;
	}

	return false;
}


const PlayerObject * getClosestTeammateToME( PlayerAgent * agent ) const
{
	const  PlayerObject * temp ;
	double distToGoal[10] = {0};
	for( int i = 1 ; i <= 10 ; i ++ )
	{
		temp = getTeammate( i + 1  , agent );
		if( temp )
			distToGoal[i-1] = temp->pos().dist(agent->world().self().pos());
		else
			distToGoal[i-1] = 1000;
	}

	int iMinDist = 1;

	for( int j = 0 ; j <= 9 ; j ++ )
		if( distToGoal[j] < distToGoal[iMinDist] )
			iMinDist = j;

	return getTeammate(iMinDist+1,agent);
}

const PlayerObject * getClosestTeammateTo( rcsc::PlayerAgent * agent , rcsc::Vector2D pos_target ) const
{
	const  PlayerObject * temp ;
	double distToTarget[10] = {0};
	for( int i = 1 ; i <= 10 ; i ++ )
	{
		temp = getOpponent( i + 1  , agent );
		if( temp )
			distToTarget[i-1] = temp->pos().dist(pos_target);
		else
			distToTarget[i-1] = 1000;
	}

	int iMinDist = 1;

	for( int j = 0 ; j <= 9 ; j ++ )
		if( distToTarget[j] < distToTarget[iMinDist] )
			iMinDist = j;

	return getOpponent(iMinDist+1,agent);
}

Triangle2D Create_Triangle_From_Two_Points( Vector2D p_r , Vector2D p_h , int a )
{
/* Graph ::

          line_h(height)
             |
             |
             |
            p_r
            /|
           / |\
          /  | \
         /   |  \
        /    |   \
       /     |    \
      /      |     \ sqrt(a^2+h^2)
     /       |h     \
    /        |       \
   /         |        \
  /          |         \
 ------------------------------- line_a
p_t_2  a/2  p_h   a/2   p_t_1

y = mx + h
                -> ( a = m , b = -1 , c = h )
ax + by + c = 0
---------------------------------

y - y0 = m*( x - x0 ) --> y = m*x - m*x0 + y0 --> y = m*x + ( y0 - m*x0 )
						  m = m , h = y0 - m*x0
*/

	double m_line_h = 0 , m_line_a = 0 , A = 0 , B = 0 , C = 0 ;
	Vector2D p_t[2];
	Circle2D circle( p_r , sqrt(pow(p_r.dist(p_h),2.0)+pow(a/2,2.0)) );

	m_line_h = (p_h.y-p_r.y)/(p_h.x-p_r.x);
	m_line_a = -1/m_line_h;
	
	A = m_line_a;
	B = -1;
	C = p_h.y - m_line_a*p_h.x;

	Line2D line(A,B,C);

	circle.intersection( line , &p_t[0] , &p_t[1] );

// 	cout<<p_r   <<endl<<
// 	      p_t[0]<<endl<<
// 	      p_t[1]<<endl<<
// 	      p_h   <<endl  ;//end test

	return Triangle2D( p_r , p_t[0] , p_t[1] );

}

Line2D getLineFromTwoPoint( Vector2D point_1 , Vector2D point_2 )
{
	double m_line_a = 0 , A = 0 , B = 0 , C = 0 ;

	m_line_a = ( point_2.y - point_1.y )/( point_2.x - point_1.x );

	A = 1;	
	B = -1*m_line_a;
	C = m_line_a*point_1.x - point_1.y;

	return Line2D(A,B,C);
}


double getRelAngleOpponentGoal( PlayerAgent * p )
{
  Vector2D posBall = p->world().ball().pos();
  Vector2D posGoal;

  if( posBall.y > 0 )
    posGoal = Vector2D( 52.5 , 7.0 );
  else
    posGoal = Vector2D( 52.5 ,-7.0 );

  return getDirection( posGoal , p->world().self().pos() ) ;
}

double getDirection( Vector2D vec1 , Vector2D vec2 )
{
	return atan2Deg( (vec1-vec2).y , (vec1-vec2).x );
}

double atan2Deg( double x , double y )
{
	if( fabs( x ) < EPSILON && fabs( y ) < EPSILON )
		return ( 0.0 );
	
	return ( Rad2Deg( atan2( x, y ) ) );
}

double Rad2Deg( AngRad x )
{
  return ( x * 180 / M_PI );
}

Vector2D getPolarPosition( const AngleDeg rel_ang , const double length = 6.0 )
{
// 	double d_x = 0 , d_y = 0;
// 	d_x = pos_start.x + length*cos(rel_ang);
// 	d_y = pos_start.y + length*sin(rel_ang);
// 	return Vector2D( d_x , d_y );
	return Vector2D::polar2vector( length , rel_ang );
}


Vector2D getRelativeOpponentGoalPos( PlayerAgent * p )
{
  Vector2D posBall = p->world().ball().pos();
  Vector2D posGoal;

  if( posBall.y > 0 )
    posGoal = Vector2D( 52.5 , 7.0 );
  else
    posGoal = Vector2D( 52.5 ,-7.0 );

	return posGoal;

}

AngleDeg getRelativeOpponentGoalAngle( PlayerAgent * p )
{
     Vector2D posBall = p->world().ball().pos();
     int sign_y( posBall.y > 0 ? 1 : -1 );
     Vector2D posGoal ( PITCH_LENGTH / 2.0 , 6.0 * sign_y );
// 
//   return Rad2Deg(atan2( (posBall.x - posGoal.x ) / (posBall.y - posGoal.y ) ) );
	return (posGoal-posBall).th();
}



AngleDeg getAgentGlobalAngle( PlayerAgent * agent )
{
	AngleDeg d_angle;
	d_angle = agent->world().self().neck();	
	return d_angle;

}


Vector2D getNearestTeammateToSelfToOpponentPenaltyArea( PlayerAgent * agent ) 
{
	const PlayerObject * nearest_tmmts[5];
	Vector2D             nearest_tmmts_pos[5] = { Vector2D(-1000,0) };

	for( int nr = 0 ; nr <= 4 ; nr ++ )
	{
		nearest_tmmts[nr] = agent->world().getTeammateNearestToSelf(10-nr);
		if( nearest_tmmts[nr] ) nearest_tmmts_pos[nr] = nearest_tmmts[nr]->pos();
	}

	for( int nr = 0 ; nr <= 4 ; nr ++ )
		if ( nearest_tmmts[nr] && nearest_tmmts[nr]->pos().x > agent->world().self().pos().x - 6.0 )
		{
		if( isInTheirPenaltyArea( agent , nearest_tmmts_pos[nr] ) && isInViewTriangle( agent , nearest_tmmts_pos[nr] ) )
			if( agent->world().self().pos().x < nearest_tmmts_pos[nr].x && agent->world().self().pos().x < 43.0 )
				return nearest_tmmts_pos[nr];
			else if( agent->world().self().pos().x > 43.0 )
				return nearest_tmmts_pos[nr];
		}

	return Vector2D(-1000,0);
}

bool isInViewTriangle( PlayerAgent * agent , Vector2D pos )
{
	const AngleDeg angle_neck = agent->world().self().neck();
	Vector2D end_pos(-1000,0);
// 	end_pos.x = agent->world().self().pos().x + 15*cos(angle_neck.degree());
// 	end_pos.y = agent->world().self().pos().x + 15*sin(angle_neck.degree());
	end_pos = getPolarPosition( angle_neck , 20.0 );

	Triangle2D triangle_view = Create_Triangle_From_Two_Points( agent->world().self().pos() , end_pos , 7 );
	return triangle_view.contains( pos );
}

int sign( double x )
{
	return ( ( x > 0 ) ? 1 : -1 );
}

int getNrInSetInHalfCircle( PlayerAgent * agent ,  Circle2D   c , bool tmm = false )
{
	const rcsc::WorldModel & wm = agent->world();// world model object
	if( !tmm )
	{
		const AbstractPlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( wm.opponent( counter + 1 ) )
			{
				players    [counter] = wm.opponent( counter + 1 );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && c.contains(players_pos[counter]) && players_pos[counter].x > wm.self().pos().x )
				nr ++;
	}
	else
	{
		const AbstractPlayerObject * players[11];
		Vector2D players_pos[11]={Vector2D(-1000,0)};
		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( wm.teammate( counter + 1 ) )
			{
				players    [counter] = wm.teammate( counter + 1 );
				players_pos[counter] = players[counter]->pos();
			}

		for( counter = 0 ; counter <= 10 ; counter ++ )
			if( players[counter] && c.contains(players_pos[counter]) )
				nr ++;
	}
	return nr;

}

/*!
	This function has decided position side.
	\brief v     :: position that has decided this side
	\return side of opponent
*/
HandSide getObjectSide( rcsc::PlayerAgent * agent , rcsc::Vector2D v )
{
	AngleDeg my_angle = agent->world().self().body();
	Vector2D my_pos   = agent->world().self().pos() ;

/*---------------------------------------------*/

	Vector2D pos_rotate(-1000,0);
	pos_rotate = v.rotate( my_angle );

	if( pos_rotate.y > v.y )
		return HS_RIGHT;
	else
		return HS_LEFT ;

/*---------------------------------------------*/
}

/*!
	This function caculate enough speed for kick ball to target.
	\brief  posPass :: target that kick to it.
	\return double form of speed model.
*/


double getEndSpeedForPass( rcsc::PlayerAgent * agent , rcsc::Vector2D posPass )
{

	double first_speed
	= rcsc::calc_first_term_geom_series_last
	( 1.8,
	agent->world().self().pos().dist( posPass ),
	rcsc::ServerParam::i().ballDecay() );

	first_speed = std::min( first_speed + 0.4,
				rcsc::ServerParam::i().ballSpeedMax() );

	return first_speed;

}

/*---------------------------------------------*/
/*!
	This function consider that first number is between two other or not.
	\brief  n1 :: Main number.
	\brief  correct form :: n2 < n1 < n3
	\return boolean Data.
*/

bool isBetween( double n1 , double n2 , double n3 )
{

	if( n1 > n2 )
		if( n1 < n3 )
			return true;

	return false;
}

/*---------------------------------------------*/
/*!
	This function consider that which of front or behind is empty.
	\return int ( 1 == BEHIND , 2 == FRONT , 0 == NOT ANY ONE ).
*/

int whichIsEmpty( rcsc::PlayerAgent * agent , rcsc::Vector2D M_home_pos )
{
	nr = 0;
	Circle2D circle_around( agent->world().self().pos() , 3.5 );
	if( getNrInSetInHalfCircle( agent , circle_around , false ) == 0 )
	{
		if( fabs(agent->world().self().pos().x - M_home_pos.x ) < 10 )
		{
			return 2;
		}
		else
			return 0;
	}
	if( getNrInSetInCircle( agent , circle_around , false ) == 0 )
	{
		if( fabs(agent->world().self().pos().x - M_home_pos.x ) < 10 )
		{
			return 1;
		}
		else
			return 0;
	}
	
	return 0;
}


int
predict_self_reach_cycle( const WorldModel & wm,
                          const Vector2D & target_point,
                          const double & dist_thr,
                          const bool save_recovery,
                          double * stamina )
{
    const ServerParam & param = ServerParam::i();
    const PlayerType & self_type = wm.self().playerType();
    const double max_moment = param.maxMoment();
    const double recover_dec_thr = param.staminaMax() * param.recoverDecThr();

    const double first_my_speed = wm.self().vel().r();

    for ( int cycle = 0; cycle < 25; ++cycle )
    {
        Vector2D my_pos = wm.self().inertiaPoint( cycle );
        double target_dist = ( target_point - my_pos ).r();
        if ( target_dist - dist_thr > self_type.realSpeedMax() * cycle )
        {
            continue;
        }

        int n_turn = 0;
        int n_dash = 0;

        AngleDeg target_angle = ( target_point - my_pos ).th();
        double my_speed = first_my_speed;

        //
        // turn
        //
        double angle_diff = ( target_angle - wm.self().body() ).abs();
        double turn_margin = 180.0;
        if ( dist_thr < target_dist )
        {
            turn_margin = std::max( 15.0,
                                    AngleDeg::asin_deg( dist_thr / target_dist ) );
        }

        while ( angle_diff > turn_margin )
        {
            angle_diff -= self_type.effectiveTurn( max_moment, my_speed );
            my_speed *= self_type.playerDecay();
            ++n_turn;
        }

        double my_stamina = wm.self().stamina();
        double my_effort = wm.self().effort();
        double my_recovery = wm.self().recovery();

        AngleDeg dash_angle = wm.self().body();
        if ( n_turn > 0 )
        {
            angle_diff = std::max( 0.0, angle_diff );
            dash_angle = target_angle;
            if ( ( target_angle - wm.self().body() ).degree() > 0.0 )
            {
                dash_angle += angle_diff;
            }
            else
            {
                dash_angle += angle_diff;
            }

            self_type.predictStaminaAfterWait( param,
                                               n_turn,
                                               &my_stamina,
                                               &my_effort,
                                               my_recovery );
        }

        //
        // dash
        //

        Vector2D vel = wm.self().vel() * std::pow( self_type.playerDecay(), n_turn );
        while ( n_turn + n_dash < cycle
                && target_dist > dist_thr )
        {
            double dash_power = std::min( param.maxPower(), my_stamina );
            if ( save_recovery
                 && my_stamina - dash_power < recover_dec_thr )
            {
                dash_power = std::max( 0.0, my_stamina - recover_dec_thr );
            }

            Vector2D accel = Vector2D::polar2vector( dash_power
                                                     * self_type.dashPowerRate()
                                                     * my_effort,
                                                     dash_angle );
            vel += accel;

            double speed = vel.r();
            if ( speed > self_type.playerSpeedMax() )
            {
                vel *= self_type.playerSpeedMax() / speed;
            }

            my_pos += vel;
            vel *= self_type.playerDecay();

            self_type.predictStaminaAfterOneDash( param,
                                                  param.maxPower(),
                                                  &my_stamina,
                                                  &my_effort,
                                                  &my_recovery );

            target_dist = my_pos.dist( target_point );
            ++n_dash;
        }

        if ( target_dist <= dist_thr )
        {
            if ( stamina )
            {
                *stamina = my_stamina;
            }
            return n_turn + n_dash;
        }
    }

    return 1000;
}


};

#endif
