// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_move.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include "body_intercept_2008.h"
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/player/say_message_builder.h>

#include "neck_offensive_intercept_neck.h"
#include "strategy.h"
#include "WorldModel.h"

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_BasicMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__": Bhv_BasicMove" );

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.95, 90.0 ).execute( agent ) )
    {
        return true;
    }

    const rcsc::WorldModel & wm = agent->world();
    WorldModelHighLevel WM;

    /*--------------------------------------------------------*/
    // chase ball
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ! wm.existKickableTeammate()
         && ( self_min <= 3
              || ( self_min < mate_min + 3
                   && self_min < opp_min + 4 )
              )
         )
    {
	//cout<<1<<wm.time().cycle()<<endl;
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": intercept" );
        rcsc::Body_Intercept2008().execute( agent );

        if ( M_turn_neck )
        {
#if 0

#else
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
#endif
        }
        return true;
    }

    const double dash_power = getDashPower( agent, M_home_pos );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    agent->debugClient().addMessage( "BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( M_home_pos );
    agent->debugClient().addCircle( M_home_pos, dist_thr );


    Vector2D move_pos = M_home_pos;
    double   dash__power = dash_power;
    if( agent->world().self().pos().dist( agent->world().ball().pos() ) < 2.0 &&
	agent->world().countTeammatesIn( Circle2D( agent->world().ball().pos() , 3.0 ) , 3 , false ) == 0 &&
	self_min < mate_min + 3 )
    {
	//cout<<2<<wm.time().cycle()<<endl;
	Body_Intercept2008().execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	return true;
    }


    bool save_stamina = true;


    if ( WM.isBetween( agent->world().ball().pos().x , -52.5 , -10.0 ) &&
         agent->world().self().stamina() > 3000 && agent->world().self().unum() <= 8 )
    {
	if( agent->world().ball().pos().x < agent->world().self().pos().x + 12 )
	{
		if( agent->world().self().stamina() > 6000 )
			save_stamina = false;
		dash__power = rcsc::ServerParam::i().maxPower();
	}
	else
	{
		save_stamina = true;
		dash__power = rcsc::ServerParam::i().maxPower() / 3.0;
	}
    }

    if ( agent->world().self().unum() >= 9 &&
         agent->world().self().pos().dist( agent->world().ball().pos() ) <= 10 &&
	 agent->world().self().stamina() > 4000 )
    {
	if( agent->world().self().stamina() > 6000 )
		save_stamina = false;
	dash__power = rcsc::ServerParam::i().maxPower();
    }

	if ( wm.ball().pos().x - wm.self().pos().x > 12.0 && 
	     WM.isBetween( wm.self().unum() , 5 , 9 ) &&
	     wm.self().stamina() < 6000 )
		dash__power *= 0.5;

	if ( wm.ball().pos().x - wm.self().pos().x < -5.0 && 
	     WM.isBetween( wm.self().unum() , 8 , 12 ) && 
	     wm.self().pos().x < wm.offsideLineX() )
		dash__power *= 0.5;

	if ( wm.ball().pos().x > 15 && self_min > mate_min )
		dash__power *= 0.5;

	if ( wm.ball().pos().x > -15.0 && wm.existKickableTeammate()  &&
	     wm.self().unum() < 7 && wm.self().pos().dist( move_pos ) < 8 &&
	     wm.ball().pos().x > wm.self().pos().x )
	{
		dash__power *= 0.5;
	}


	if ( move_pos == wm.ball().pos() &&
	     self_min < mate_min + 2 )
	{
		//cout<<3<<wm.time().cycle()<<endl;
		Bhv_GoToPointLookBall( wm.ball().pos(),
				       dist_thr,
				       rcsc::ServerParam::i().maxPower()
				     ).execute( agent );
		return true;
	}


    if ( ! rcsc::Body_GoToPoint( move_pos , dist_thr, dash__power , 1 , false , save_stamina , 15
                                 ).execute( agent ) )
    {
	//cout<<4<<wm.time().cycle()<<endl;
        if ( M_turn_at )
        {
            rcsc::Body_TurnToBall().execute( agent );
        }
        else
        {
            return false;
        }
    }

    if ( M_turn_neck )
    {
        if ( wm.existKickableOpponent()
             && wm.ball().distFromSelf() < 18.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
    }
//     agent->setArmAction( new Arm_PointToPoint( move_pos ) );
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_BasicMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & /*target_point*/ )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();

    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    // check recover
    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": getDashPower() correct DF line. keep max power" );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": getDashPower() recovering" );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": getDashPower() exist kickable teammate. dash_power=%.1f",
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": in offside area. dash_power=%.1f",
                            dash_power );
    }
    else if ( wm.ball().pos().x > 25.0
              && wm.ball().pos().x > wm.self().pos().x > 10.0
              && self_min < opp_min - 6
              && mate_min < opp_min - 6 )
    {
        dash_power = rcsc::bound( rcsc::ServerParam::i().maxPower() * 0.1,
                                  my_inc * 0.5,
                                  rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": getDashPower() opponent ball dash_power=%.1f",
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": normal mode dash_power=%.1f",
                            dash_power );
    }

	if( agent->world().self().unum() > 8 )
	{
		if( agent->world().self().stamina() < 5000 &&
		    agent->world().self().pos().dist( agent->world().ball().pos() ) > 18 )
		{
			return rcsc::ServerParam::i().maxPower()/3.0;
		}
		else
			switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
				case Strategy::BA_CrossBlock:
				{
					return rcsc::ServerParam::i().maxPower()/4.0;
				}
				case Strategy::BA_Stopper:
				{
					if( agent->world().ball().pos().x > -10.0 )
						return rcsc::ServerParam::i().maxPower()/2.5;
					else
						 return rcsc::ServerParam::i().maxPower()/4.0;
				}
				case Strategy::BA_Danger:
				{
					return rcsc::ServerParam::i().maxPower()/5.0;
				}
				case Strategy::BA_DribbleBlock:
				{
					if( agent->world().ball().pos().x > -10.0 )
						return rcsc::ServerParam::i().maxPower()/2.5;
					else
						 return rcsc::ServerParam::i().maxPower()/4.0;
				}
				case Strategy::BA_DefMidField:
				{
					if( agent->world().ball().pos().x > -10.0 )
						return rcsc::ServerParam::i().maxPower()/2.0;
					else
						 return rcsc::ServerParam::i().maxPower()/3.0;
				}
				case Strategy::BA_DribbleAttack:
				{
					return rcsc::ServerParam::i().maxPower();
				}
				case Strategy::BA_OffMidField:
				{
					return rcsc::ServerParam::i().maxPower();
				}
				case Strategy::BA_Cross:
				{
					return rcsc::ServerParam::i().maxPower();
				}
				case Strategy::BA_ShootChance:
				{
					return rcsc::ServerParam::i().maxPower();
				}
				default:
					return rcsc::ServerParam::i().maxPower();
				break;
			}
	}

	if( wm.ball().pos().dist( wm.self().pos() ) > 22.0 &&
	    wm.ball().pos().x > -10.0 )
		return rcsc::ServerParam::i().maxPower() / 2.5;

	if( fabs( agent->world().ball().pos().x - agent->world().self().pos().x ) > 20 )
		return rcsc::ServerParam::i().maxPower() / 2;

	if( wm.ball().pos().x < 1.0 && wm.self().unum() < 7 )
		return rcsc::ServerParam::i().maxPower();


	if( wm.self().isKickable() || wm.ball().pos().x < 10.0 ||
	  ( wm.self().unum() >= 7  && wm.ball().pos().x > 30.0 ))
		return rcsc::ServerParam::i().maxPower();


    return dash_power;
}
