#include "bhv_dribble.h"

#include <rcsc/action/body_dribble.h>
#include <rcsc/action/obsolete/body_kick_multi_step.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/body_dribble2008.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_intercept2008.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "bhv_penalty_kick.h"
#include "bhv_dribble2009.h"
#include "bhv_cross_to_center.h"
#include "body_kick_to_corner.h"
#include "WorldModel.h"
#include "move_to_pos.h"
#include "bhv_basic_offensive_kick.h"


#define PITCH_WIDTH  68.0
#define PITCH_LENGTH 105.0


/*!
  executable Dribble.
*/

bool Bhv_Dribble::execute( rcsc::PlayerAgent * agent )
{

	double        dLength;

	const rcsc::Vector2D  posAgent = agent->world().self().pos();
	const AngleDeg        angBody = WM.getAgentGlobalAngle( agent );
	
	ang = getDribbleAngle( agent );
	dribbleT = getDribbleType( agent );


	const PlayerObject * nearest_opp = agent->world().getOpponentNearestToBall(10);
	Vector2D nearest_opp_pos(-1000,0);
	if( nearest_opp ) nearest_opp_pos = agent->world().getOpponentNearestToBall(10)->pos();

	switch( dribbleT )
	{
		case DRIBBLE_WITHBALL:
			dLength = 7.5;
			break;
		case DRIBBLE_SLOW:
			dLength = 15.0;
			break;
		case DRIBBLE_FAST:
			dLength = 22.5;
			break;
		default:
			dLength = 0.0;
			break;
	}

	rcsc::Vector2D Move_Pos(-1000,0);

	Move_Pos = Vector2D::polar2vector( dLength + 15 , ang  );

	bool dodge = true;
	if ( agent->world().self().pos().x > 43.0
		&& agent->world().self().pos().absY() < 20.0 )
	{
		dodge = false;
	}

	int dash_count = 10;
	const rcsc::Rect2D safety_rect
		= rcsc::Rect2D::from_center( agent->world().ball().pos().x + 6.5,
						agent->world().ball().pos().y      ,
						13.0                               ,
						13.0                              );

	if ( agent->world().existOpponentIn( safety_rect, 10, false ) )
	{
		dash_count = 3;
	}

	rcsc::Vector2D posDribble = posAgent + Move_Pos;

	if( posAgent.x > 50 || posAgent.x < - 48 )
		posDribble.x = 48 * sign( posAgent.x ) ;
	if( posAgent.y > 30 || posAgent.y < -30 )
		posDribble.y = 25 * sign( posAgent.y );


	const rcsc::WorldModel & wm = agent->world();
	rcsc::Triangle2D triangle( wm.self().pos() ,
					wm.self().pos() + rcsc::Vector2D(20.0, 15.0),
					wm.self().pos() + rcsc::Vector2D(20.0,-15.0));
	const double nearest_opp_dist = ( nearest_opp
					? nearest_opp->distFromSelf()
					: 1000.0 );

	const PlayerPtrCont::const_iterator opp_to_ball = wm.opponentsFromBall().begin();


	if( nearest_opp && nearest_opp_dist <= ServerParam::i().tackleDist() + 0.2 )
	{

		if( posAgent.x > nearest_opp_pos.x - 1 || fabs(posAgent.y - nearest_opp_pos.y) > 1.0 )
		{
			if( wm.self().unum() > 8 &&
				wm.countOpponentsIn( triangle , 10 , false ) == 0 &&
				posAgent.absX() < 48 && posAgent.absY() < 30  )
			{
				rcsc::Body_SmartKick( posDribble ,
						1.1 ,
						1.0,
						3 ).execute( agent );
			}
			else
			{
				Bhv_BasicOffensiveKick().execute( agent );
			}
		}
		else
		{
			if( posAgent.x > 38.0 )
			{
				if( posAgent.absY() > 17 )
				{
					Vector2D cross_to_center_point( 52.5 , 0.0 );
					bool is_let_to_cross = Bhv_CrossToCenter().get_best_point( agent , &cross_to_center_point );
					if( is_let_to_cross )
					{
						AngleDeg cross_to_center_ang = ( cross_to_center_point - posAgent ).th();
						double   dist_to_center      = posAgent.dist( cross_to_center_point );
						Sector2D sector_to_center( posAgent,
									0.5,
									dist_to_center,
									cross_to_center_ang + 20.0,
									cross_to_center_ang - 20.0 );
	
						if( !wm.existOpponentIn( sector_to_center , 10 , false ) &&
						wm.existTeammateIn( sector_to_center , 10 , false ) &&
						posAgent.absY() > 15 )
						{
							Bhv_CrossToCenter().execute( agent );
							return true;
						}
					}
					else
					{
						if( Body_Pass().execute( agent ) )
						{
							agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
							return true;
						}
						else
						{
							Bhv_Dribble2009().execute( agent );
							agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
							return true;
						}
					}
				}
				else
				{
					Vector2D kick_to_corner_point = Body_KickToCorner( posAgent.y < 0
											).kick_to_corner_target();
					AngleDeg kick_to_corner_ang = ( kick_to_corner_point - posAgent ).th();
	
					double dist_to_corner = posAgent.dist( kick_to_corner_point );
	
					Sector2D sector_to_corner ( posAgent,
								0.5,
								dist_to_corner,
								kick_to_corner_ang + 20,
								kick_to_corner_ang - 20 );
	
					if( !wm.existOpponentIn( sector_to_corner , 10 , false ) &&
					wm.existTeammateIn( sector_to_corner , 10 , false ) &&
					posAgent.absY() < 15 )
					{
						Body_KickToCorner( posAgent.y < 0 ).execute( agent );
						agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
						return true;
					}
					else
					{
						if( Body_Pass().execute( agent ) )
						{
							agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
							return true;
						}
						else
						{
							Bhv_Dribble2009().execute( agent );
							agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
							return true;
						}
					}
				}
			}
			else
			{
				Body_Dribble2008( posDribble,
							1.0,
							rcsc::ServerParam::i().maxDashPower(),
							dash_count,
							dodge
						).execute( agent );
			}
		}
	}
	else
	{
		if( !nearest_opp ||
			( nearest_opp && wm.countOpponentsIn( triangle , 10 , false ) == 0 &&
			nearest_opp_dist > 4.0 && posAgent.x > nearest_opp_pos.x  &&
			( posAgent.x < 32.0 || ( posAgent.x < 35 && fabs(posAgent.y) > 20.0 ) ) &&
			agent->world().self().unum() > 8 && agent->world().self().unum() < 11 ) )
		{
			rcsc::Body_SmartKick( posAgent + Vector2D( 3.0, 0.0 ),
					1.2,
					1.0,
					3 ).execute( agent );
		}
		else if( posAgent.x < 41 )
		{
			Body_Dribble2008( posDribble,
					1.2,
					rcsc::ServerParam::i().maxDashPower(),
					dash_count,
					dodge
					).execute( agent );
		}
		else
		{
			Vector2D kick_to_corner_point = Body_KickToCorner( posAgent.y < 0
										).kick_to_corner_target();
			AngleDeg kick_to_corner_ang = ( kick_to_corner_point - posAgent ).th();

			double dist_to_corner = posAgent.dist( kick_to_corner_point );

			Sector2D sector_to_corner ( posAgent,
							0.5,
							dist_to_corner,
							kick_to_corner_ang + 20,
							kick_to_corner_ang - 20 );

			if( !wm.existOpponentIn( sector_to_corner , 10 , false ) &&
				wm.existTeammateIn( sector_to_corner , 10 , false ) &&
				posAgent.absY() < 15 )
			{
				Body_KickToCorner( posAgent.y < 0 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
				return true;
			}

			Vector2D cross_to_center_point( 52.5 , 0.0 );
			bool is_let_to_cross = Bhv_CrossToCenter().get_best_point( agent , &cross_to_center_point );
			if( is_let_to_cross )
			{
				AngleDeg cross_to_center_ang = ( cross_to_center_point - posAgent ).th();
				double   dist_to_center      = posAgent.dist( cross_to_center_point );
				Sector2D sector_to_center( posAgent,
								0.5,
								dist_to_center,
								cross_to_center_ang + 20.0,
								cross_to_center_ang - 20.0 );

				if( !wm.existOpponentIn( sector_to_center , 10 , false ) &&
					wm.existTeammateIn( sector_to_center , 10 , false ) &&
					posAgent.absY() > 15 )
				{
					Bhv_CrossToCenter().execute( agent );
					agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
					return true;
				}
			}

			dribble( agent , -90*sign( posAgent.y ) , DRIBBLE_WITHBALL , 10 );
			agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
			return true;
		}
	}

	agent->setNeckAction( new rcsc::Neck_TurnToPoint( wm.ball().pos() ) );
	return true;
}

/*!
   consider dribble sign.
*/

int Bhv_Dribble::sign( double x )
{
	return ( ( x > 0 ) ? 1 : -1 );
}


/*!
   consider dribble angle.
*/


double Bhv_Dribble::getDribbleAngle( PlayerAgent * agent )
{
	Vector2D my_pos = agent->world().self().pos();

	const PlayerObject * nearest_opp = agent->world().getOpponentNearestToSelf(10);

	Vector2D nearest_opp_pos(-1000,0);

	if( nearest_opp ) nearest_opp_pos = nearest_opp->pos();

	if( my_pos.x >= 51.0 )
	{
		if( my_pos.y > 0 )
		{
			return -90;
		}
		else
		{
			return  90;
		}
	}

	if( my_pos.x >= 35.0 )
	{
		if( my_pos.y > 0 )
		{
			return -45;
		}
		else
		{
			return  45;
		}
	}

	if( nearest_opp
	    && my_pos.x >= nearest_opp_pos.x
	    && fabs(my_pos.y) > 15 )
	{
		return 0.0;
	}

	if( WM.isInTheirPenaltyArea( agent , my_pos ) )
	{
		return WM.getRelAngleOpponentGoal( agent );
	}

	if( nearest_opp && my_pos.x >= nearest_opp_pos.x )
	{
		return 0.0;
	}

	if( nearest_opp && my_pos.y > nearest_opp_pos.y )
	{
		return  45.0;
	}

	if( nearest_opp && my_pos.y < nearest_opp_pos.y )
	{
		return -45.0;
	}

	return 0;
}


/*!
   consider dribble type.
*/

DribbleT Bhv_Dribble::getDribbleType( PlayerAgent * agent )
{
	Vector2D my_pos = agent->world().self().pos();

	const PlayerObject * nearest_opp = agent->world().getOpponentNearestToSelf(10);

	Vector2D nearest_opp_pos(-1000,0);

	AngleDeg angle_me_to_goal = WM.getRelativeOpponentGoalAngle( agent ).degree();
	

	if( agent->world().self().stamina() < 2500 )
	{
	   return DRIBBLE_WITHBALL;
	}

	if( my_pos.x > 45.0 )
	{
	   return DRIBBLE_SLOW;
	}

	if( nearest_opp )
	{
	   nearest_opp_pos = agent->world().getOpponentNearestToSelf(10)->pos();
	}
	
	Triangle2D triangle_me_to_goal
	= WM.Create_Triangle_From_Two_Points( my_pos - Vector2D(1.5,0.0) ,
						WM.getPolarPosition( angle_me_to_goal , 20 ) , 9 );
	
	Triangle2D triangle_me_to_front
	= WM.Create_Triangle_From_Two_Points(  my_pos - Vector2D(1.5,0.0) ,
						WM.getPolarPosition( 0 , 15 ) , 9 );
	
	if( !nearest_opp )
	{
	   return DRIBBLE_FAST;
	}

	if( nearest_opp && my_pos.dist( nearest_opp_pos ) <= 0.1 )
	{
	   return DRIBBLE_SLOW;
	}

	if( WM.getNrInSetInTriangle( agent , triangle_me_to_goal ) <= 1 )
	{
		if( my_pos.dist(Vector2D ( PITCH_LENGTH / 2.0 , 0.0 ) ) > 15 )
		{
		   return DRIBBLE_FAST;
		}
		else
		{
		   return DRIBBLE_SLOW;
		}
	}

	if( WM.getNrInSetInTriangle( agent , triangle_me_to_front ) == 0 )
	{
		if( nearest_opp && nearest_opp_pos.dist( my_pos ) >= 7.5 )
		{
		   return DRIBBLE_FAST;
		}
		else
		{
		   return DRIBBLE_SLOW;
		}
	}

	if( nearest_opp
	    && my_pos.x > nearest_opp_pos.x
	    && nearest_opp
	    && nearest_opp_pos.dist( my_pos ) >= 7.5 )
	{
	   return DRIBBLE_FAST;
	}
	else if( nearest_opp && my_pos.x > nearest_opp_pos.x )
	{
	   return DRIBBLE_FAST;
	}
	else
	{
	   return DRIBBLE_SLOW;
	}

	return DRIBBLE_WITHBALL;
}

/*!


*/

void Bhv_Dribble::dribble( PlayerAgent * agent ,
                           double angle,
                           DribbleT DT,
                           int numberDash )
{
	double dLength =0;

	switch( DT )
	{
	case DRIBBLE_WITHBALL:
		dLength = 5.0;
		break;
	case DRIBBLE_SLOW:
		dLength = 9.0;
		break;
	case DRIBBLE_FAST:
		dLength = 14.0;
		break;
	default:
		dLength = 0.0;
		break;
	}
	Vector2D posDribble = agent->world().self().pos() + WM.getPolarPosition( ang , dLength );
	if( posDribble.x > 50 )
	{
		posDribble.x = 50;
	}
	
	if( posDribble.absY() > 32 )
	{
		posDribble.y = 32 * WM.sign( posDribble.y );
	}

	rcsc::Body_Dribble2008(  posDribble , 3.5 , 100 , numberDash , 1 ).execute( agent );
	return;
}


bool Bhv_Dribble::doLongDribble( rcsc::PlayerAgent * agent )
{
		
	const rcsc::WorldModel & wm = agent->world();
	
	// dribble Long
	if ( wm.self().pos().x > 36.0 )
	{
		return false;
	}
	
	if ( wm.self().pos().absY() > rcsc::ServerParam::i().pitchHalfWidth() - 2.0 )
	{
		if ( wm.self().pos().y < 0.0 && wm.self().body().degree() < 1.5 )
		{
		return false;
		}
		if ( wm.self().pos().y > 0.0 && wm.self().body().degree() > -1.5 )
		{
		return false;
		}
	}
	
	const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
	if ( nearest_opp
		&& nearest_opp->pos().x > wm.self().pos().x + 0.5
		&& std::fabs( nearest_opp->pos().y -  wm.self().pos().y ) < 2.0
		&& nearest_opp->distFromSelf() < 2.0
		)
	{
		return false;
	}
	
	const rcsc::Rect2D target_rect
		= rcsc::Rect2D::from_center( wm.ball().pos().x + 5.5,
					wm.ball().pos().y,
					10.0, 9.0 );
	if ( wm.existOpponentIn( target_rect, 10, false ) )
	{
		return false;
	}
	
	const rcsc::Rect2D safety_rect
		= rcsc::Rect2D::from_center( wm.ball().pos().x + 6.5,
					wm.ball().pos().y,
					13.0, 13.0 );
	//int dash_count = 3;
	int dash_count = 10;
	if ( wm.existOpponentIn( safety_rect, 10, false ) )
	{
		//dash_count = 1;
		dash_count = 8;
	}
	
	rcsc::Vector2D drib_target( 47.0, wm.self().pos().y );
	if ( wm.self().body().abs() < 20.0 )
	{
		drib_target
		= wm.self().pos()
		+ rcsc::Vector2D::polar2vector( 10.0, wm.self().body() );
	}
	
	agent->debugClient().addMessage( "LongDrib" );
	agent->debugClient().addRectangle( target_rect );
	
	rcsc::Body_Dribble2008( drib_target,
				1.0,
				rcsc::ServerParam::i().maxPower(),
				dash_count,
				false // no dodge
				).execute( agent );
	
	const rcsc::PlayerObject * opponent = static_cast< rcsc::PlayerObject * >( 0 );
	
	const rcsc::PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( rcsc::PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		if ( (*it)->distFromSelf() > 10.0 ) break;
		if ( (*it)->posCount() > 5 ) continue;
		if ( (*it)->posCount() == 0 ) continue;
		if ( (*it)->angleFromSelf().abs() > 90.0 ) continue;
		if ( wm.dirCount( (*it)->angleFromSelf() ) < (*it)->posCount() ) continue;
	
		// TODO: check the next visible range
	
		opponent = *it;
		break;
	}
	
	if ( opponent )
	{
		agent->debugClient().addMessage( "LookOpp" );
		agent->setNeckAction( new rcsc::Neck_TurnToPoint( opponent->pos() ) );
	}
	else
	{
		agent->setNeckAction( new rcsc::Neck_ScanField() );
	}

	return true;
}
