// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "bhv_dribble2009.h"

#include <rcsc/action/body_dribble2008.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include "WorldModel.h"

bool Bhv_Dribble2009::execute( rcsc::PlayerAgent * agent )
{
	const rcsc::WorldModel & wm = agent->world();
	rcsc::Vector2D posTarget(-1000,0);
	rcsc::Vector2D posAgent = wm.self().pos();
	const rcsc::PlayerObject * closest_opp = wm.getOpponentNearestToSelf( 10 );
	rcsc::Vector2D closest_opp_pos( -1000 , 0 );
	if( closest_opp ) closest_opp_pos = closest_opp->pos();
	double closest_opp_dist = ( closest_opp ? closest_opp_pos.dist( posAgent ) : -100 );

	int iSign = 1;
	if( posAgent.y < 0 )iSign=-1;

	rcsc::Sector2D sector_l_s ( posAgent , 0.5 , 10.0 ,-60 , -20 );
	rcsc::Sector2D sector_c   ( posAgent , 0.5 , 10.0 ,-20 ,  20 );
	rcsc::Sector2D sector_r_s ( posAgent , 0.5 , 10.0 , 20 ,  60 );

	if( posAgent.x > 36.5 )
	{
		WorldModelHighLevel WM;
		rcsc::AngleDeg diff_a = ( rcsc::Vector2D( 52.5 , 0.0 ) - posAgent ).th();
		rcsc::Sector2D sector_move_p( posAgent , 0.5 , 6.0 ,diff_a - 20,diff_a + 20 );
		
		if( wm.countOpponentsIn( sector_move_p , 6 , true ) == 0 )
			posTarget = rcsc::Vector2D( 52.5 , 0.0 );
		else
		{
			if( wm.self().pos().x < 48.0 )
			{
				posTarget = rcsc::Vector2D( 52.5 , 0.0 );
			}
			else
				posTarget = rcsc::Vector2D(-4.0, -4.0*iSign );
		}
	}
	else if( closest_opp && closest_opp_pos.x < posAgent.x - 3 )
	{
		posTarget = rcsc::Vector2D( 52.5 , 0.0 );
	}
	else
	{
		if( iSign )
		{
			if( wm.countOpponentsIn( sector_c , 10 , false ) == 0 &&
				closest_opp_dist > 5 )
			{
				posTarget = rcsc::Vector2D( 10 , 0 );
			}
			else if( wm.countOpponentsIn( sector_r_s , 10 , false ) == 0 )
			{
				posTarget = rcsc::Vector2D( 10 , 12 );
			}
			else  if( wm.countOpponentsIn( sector_l_s , 10 , false ) == 0 )
			{
				posTarget = rcsc::Vector2D( 10 , -12 );
			}
			else
			{
				posTarget = rcsc::Vector2D( -10 , 0 );
			}
		}
		else
		{
			if( wm.countOpponentsIn( sector_c , 10 , false ) == 0 &&
				closest_opp_dist > 5 )
			{
				posTarget = rcsc::Vector2D( 10 , 0 );
			}
			else if( wm.countOpponentsIn( sector_l_s , 10 , false ) == 0 )
			{
				posTarget = rcsc::Vector2D( 10  ,-12 );
			}
			else  if( wm.countOpponentsIn( sector_r_s , 10 , false ) == 0 )
			{
				posTarget = rcsc::Vector2D( 10 , 12 );
			}
			else
			{
				posTarget = rcsc::Vector2D( -10 , 0 );
			}
		}
	}

	AngleDeg move_deg = posTarget.th();

	cout<<"BEST DRIBBLE in cycle :: "<<agent->world().time().cycle()<<endl;


	if( ( posAgent + posTarget ).absY() > 30 )
		posTarget.y = 0;

	rcsc::Sector2D sector_move( posAgent - Vector2D( 3.0 , 0.0 ),
					0.5 ,
					10.0,
					move_deg - 40,
					move_deg + 40 );

	rcsc::Circle2D circle_round( posAgent , 7 );
	WorldModelHighLevel WM2;

	if( posAgent.absX() > 50 )
	{
		posTarget.x = 50 * sign( posTarget.x );
	}
	if( posAgent.absY() > 32 )
	{
		posTarget.y = 32 * sign( posTarget.y );
	}


	if( wm.countOpponentsIn( sector_move  , 15 , false ) == 0 &&
		WM2.getNrInSetInHalfCircle( agent , circle_round ) == 0 &&
		wm.self().pos().x < 36.0 )
	{
		rcsc::Body_SmartKick( posAgent + posTarget / 5 ,
					0.9,
					0.8,
					2 ).execute( agent );
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
		return true;
	}

	if( wm.countOpponentsIn( sector_move  , 15 , false ) == 0 )
		rcsc::Body_Dribble2008( posAgent + rcsc::Vector2D ( posTarget.x , posTarget.y) ,
					1.0 ,
					rcsc::ServerParam::i().maxDashPower(),
					3,
					false ).execute( agent );
	else
		Body_Pass().execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	return true;

}
