// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher.

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef BHV_GOALIE_DANGER_AREA_MOVE_H
#define BHV_GOALIE_DANGER_AREA_MOVE_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/line_2d.h>

#include <boost/shared_ptr.hpp>


#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/soccer_math.h>

#include <rcsc/action/view_wide.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include "body_intercept_2008.h"
#include "neck_default_intercept_neck.h"

namespace rcsc {
class Formation;
}

class Bhv_GoalieDangerAreaMove
    : public rcsc::SoccerBehavior {
public:

    Bhv_GoalieDangerAreaMove()
      { }

    bool execute( rcsc::PlayerAgent * agent )
    {
	rcsc::Vector2D pos_target = getTargetPoint( agent );

	if( pos_target.valid() )
	{
		if( pos_target == agent->world().ball().pos() )
		{
			rcsc::Body_Intercept2008().execute( agent );
			agent->setNeckAction( new Neck_DefaultInterceptNeck( new rcsc::View_Wide() , 
									     new rcsc::Neck_TurnToBallOrScan() ) );
			return true;
		}
		else if( pos_target.dist( agent->world().ball().pos() ) < 2.0 )
		{
			rcsc::Bhv_GoToPointLookBall( pos_target,
						     0.5,
						     rcsc::ServerParam::i().maxDashPower() ).execute( agent );
			return true;
		}
		else
		{
			int cycle = int ( agent->world().self().pos().dist( agent->world().ball().pos() ) / 3.0 );
			if( !rcsc::Body_GoToPoint( pos_target,
						   cycle,
						   true,
						   false,
						   18.0 ).execute( agent ) )
			{
				rcsc::Body_TurnToBall().execute( agent );
			}

			agent->setNeckAction( new rcsc::Neck_ScanField() );

			return true;
		}
	}

	return false;
    }

private:

    rcsc::Vector2D getTargetPoint( rcsc::PlayerAgent * agent )
    {

	const rcsc::WorldModel & wm = agent->world();

	rcsc::Vector2D ball_pos  = wm.ball().pos();
	rcsc::Vector2D ball_vel  = wm.ball().vel();
	rcsc::Line2D   ball_line = rcsc::Line2D( ball_pos , ball_pos + ball_vel );

	rcsc::Vector2D agent_pos = wm.self().pos();
	rcsc::Line2D   agent_line= rcsc::Line2D( agent_pos , agent_pos + rcsc::Vector2D( 0.0 , 1.0 ) );

	rcsc::Vector2D intersect_pos = rcsc::Line2D::intersection( ball_line , agent_line );

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min  = wm.interceptTable()->opponentReachCycle();


	if( self_min < mate_min && self_min < opp_min + 3 )
	{
		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" ball is nearest to me and return ball pos now." );
		return ball_pos;
	}


	if( !intersect_pos.valid() )
	{
		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" intersection pos is invalid and return INVALIDATED vector." );
		return rcsc::Vector2D::INVALIDATED;
	}

	if( intersect_pos.absY() < 7.01 && !wm.existKickableOpponent() &&
	    ball_vel.r() > 1.8 )
	{
		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" ball is heading to ball and return to intersection pos." );
		return intersect_pos;
	}

	return rcsc::Vector2D::INVALIDATED;
    }


};

#endif
