// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "bhv_mark_close_opponent.h"

static bool _S_stop_move;

bool Bhv_MarkCloseOpponent::execute( rcsc::PlayerAgent * agent )
{
	const PlayerObject * closest_opp = agent->world().getOpponentNearestToSelf(10);
	Vector2D         closest_opp_pos( 1000 , 0 );
	if( closest_opp ) closest_opp_pos = closest_opp->pos();

	if( closest_opp )
	{
		double min_dist = 10.0;
		const PlayerObject * tmm_nearest_to_close_opp =
		      agent->world().getTeammateNearestTo( closest_opp_pos , 10, &min_dist );
		Vector2D tmm_nearest_to_close_opp_pos ( -1000 , 0  );

		if( tmm_nearest_to_close_opp ) tmm_nearest_to_close_opp_pos = tmm_nearest_to_close_opp->pos();

		if( tmm_nearest_to_close_opp )
		{
			double tmm_nearest_to_close_opp_dist = 
			       tmm_nearest_to_close_opp_pos.dist( closest_opp_pos );

			double self_dist = 
			       agent->world().self().pos().dist ( closest_opp_pos );


			if( self_dist < tmm_nearest_to_close_opp_dist - 2 &&
			    !Circle2D( closest_opp_pos , 2   ).contains( tmm_nearest_to_close_opp_pos ) &&
			    !Circle2D( closest_opp_pos , 8.5 ).contains( agent->world().ball().pos() ))
			{
				const WorldModel & wm = agent->world();
				Vector2D mark_point = closest_opp_pos;
				Vector2D extra_point = get_extra_point( agent , mark_point );
				double dist_thr = wm.self().pos().dist(mark_point+extra_point)*0.1;


				if( wm.self().pos().dist( mark_point + extra_point * 2 ) < 0.5 )
				{
					_S_stop_move = false;
				}
				else
				{
					_S_stop_move = true;
				}


				if( wm.self().unum() < 6 && mark_point.x > 0.0 )
				{
					return false;
				}
	
				if( wm.ball().pos().absY() < 28 || wm.ball().pos().x > -30.0 )
				{
					if( Body_GoToPoint2009( mark_point + extra_point * 2,
								dist_thr,
								ServerParam::i().maxPower() ).execute( agent )
								&& _S_stop_move )
					{
						agent->setNeckAction( new rcsc::Neck_ScanField() );
						return true;
					}
					else
					{
						Body_TurnToPoint( closest_opp_pos ).execute( agent );
						return true;
					}
				}
				else
					return false;
			}
			else
			{
				const WorldModel & wm = agent->world();
				WorldModelHighLevel WM;
				PlayerObject * opp = NULL;
				const PlayerPtrCont::const_iterator o_beg = wm.opponentsFromSelf().begin();
				const PlayerPtrCont::const_iterator o_end = wm.opponentsFromSelf().end  ();

				opp = *(o_beg + 1);

				if( !opp )
					return false;

				Vector2D mark_point = opp->pos();
				double min_dist = 5.0;

				const PlayerObject * near_to_mark =
				      wm.getTeammateNearestTo( mark_point , 10 , &min_dist );
				if( !near_to_mark )
				{

					const WorldModel & wm = agent->world();
					Vector2D extra_point = get_extra_point( agent , mark_point );
					double dist_thr = wm.self().pos().dist(mark_point+extra_point)*0.1;

					if( wm.self().pos().dist( mark_point + extra_point * 2 ) < 0.5 )
					{
						_S_stop_move = false;
					}
					else
					{
						_S_stop_move = true;
					}

					if( wm.ball().pos().absY() < 28 || wm.ball().pos().x > -30.0 )
					{
						if( Body_GoToPoint2009( mark_point + extra_point * 2,
									dist_thr,
									ServerParam::i().maxPower()
								      ).execute( agent )
									&& _S_stop_move )
						{
							agent->setNeckAction( new rcsc::Neck_ScanField() );
							return true;
						}
						else
						{
							Body_TurnToPoint( closest_opp_pos ).execute( agent );
							return true;
						}
					}
					else
						return false;
				}
				else
				{
				    if( self_dist < tmm_nearest_to_close_opp_dist - 2 &&
				        !Circle2D(closest_opp_pos,3).contains( tmm_nearest_to_close_opp_pos ) 
					&& !Circle2D( closest_opp_pos , 8.5 ).contains(
						      agent->world().ball().pos()) )
					{
	
						const WorldModel & wm = agent->world();
						Vector2D extra_point = get_extra_point( agent , mark_point );
						double dist_thr =
						       wm.self().pos().dist(mark_point+extra_point)*0.1;
	
						if(wm.self().pos().dist(mark_point + extra_point * 2) < 0.5)
						{
							_S_stop_move = false;
						}
						else
						{
							_S_stop_move = true;
						}
	
						if( wm.ball().pos().absY() < 28 || wm.ball().pos().x > -30.0 )
						{
							if( Body_GoToPoint2009( mark_point + extra_point * 2,
										dist_thr,
										ServerParam::i().maxPower()
									).execute( agent )
										&& _S_stop_move )
							{
								agent->setNeckAction(
								       new rcsc::Neck_ScanField() );
								return true;
							}
							else
							{
								Body_TurnToPoint(closest_opp_pos
										).execute( agent );
								return true;
							}
						}
						else
							return false;
					}
				}

				return false;
			}
		}
/////////////////////////////////////////////////////////////
		else if( !Circle2D( closest_opp_pos , 2 ).contains( tmm_nearest_to_close_opp_pos ) )
		{
			const WorldModel & wm = agent->world();
			Vector2D mark_point = closest_opp_pos;
			Vector2D extra_point = get_extra_point( agent , mark_point );
			double dist_thr = wm.self().pos().dist(mark_point+extra_point)*0.1;

			if( wm.self().pos().dist( mark_point + extra_point * 2 ) < 0.5 )
			{
				_S_stop_move = false;
			}
			else
			{
				_S_stop_move = true;
			}

			if( wm.self().unum() < 6 && mark_point.x > 0.0 )
			{
				return false;
			}

			if( wm.ball().pos().absY() < 28 || wm.ball().pos().x > -30.0 )
			{
			
				if( Body_GoToPoint2009( mark_point + extra_point * 2,
							dist_thr,
							ServerParam::i().maxPower() ).execute( agent )
							&& _S_stop_move )
				{
					agent->setNeckAction( new rcsc::Neck_ScanField() );
					return true;
				}
				else
				{
					Body_TurnToPoint( closest_opp_pos ).execute( agent );
					return true;
				}
			}
			else
				return false;
			return true;
		}
 /////////////////////////////////////////////////////////////

	}
	else
		return false;

	return false;
}


Vector2D Bhv_MarkCloseOpponent::get_extra_point( PlayerAgent * agent , Vector2D mark_point )
{
	const WorldModel & wm = agent->world();
	Vector2D extra_point( 0.0 , 0.0 );
	double diff = mark_point.x - wm.ball().pos().x;

	if( wm.ball().pos().y > 28 )
	{
		if( diff > 7 )
		{
			extra_point.assign( -1.0 , 0.0 );
		}
		else if( diff < -7 )
		{
			extra_point.assign( 1.0 , 0.0 );
		}
		else
		{
			extra_point.assign( 0.0 , 1.0 );
		}
	}
	else if( wm.ball().pos().y < -28 )
	{
		if( diff > 7 )
		{
			extra_point.assign( -1.0 , 0.0 );
		}
		else if( diff < -7 )
		{
			extra_point.assign( 1.0 , 0.0 );
		}
		else
		{
			extra_point.assign( 0.0 ,-1.0 );
		}
	}
	else if( diff > 0 )
	{
		double diff_y = mark_point.y - wm.ball().pos().y ;
		if( diff_y > 5 )
		{
			extra_point.assign( -0.7 , -0.7 );
		}
		else if( diff_y < -5 )
		{
			extra_point.assign( -0.7 ,  0.7 );
		}
		else
		{
			extra_point.assign( -1.0 , 0.0 );
		}
	}
	else if( diff <= 0 )
	{
		double diff_y = mark_point.y - wm.ball().pos().y ;
		if( diff_y > 5 )
		{
			extra_point.assign( 0.7 , -0.7 );
		}
		else if( diff_y < -5 )
		{
			extra_point.assign( 0.7 ,  0.7 );
		}
		else
		{
			extra_point.assign( 1.0 , 0.0 );
		}
	}
	else
	{
		extra_point.assign( 0 , 0 );
	}

	return extra_point;

}
