// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "bhv_mark_opponent.h"

bool Bhv_MarkOpponent::execute( PlayerAgent * agent )
{
        if( S_last_mark_time != agent->world().time().cycle() &&
            fabs( S_last_mark_time - agent->world().time().cycle() ) > 200 )
        {
           S_last_mark_time = agent->world().time().cycle();
        }

        if( agent->world().time().cycle() - S_last_mark_time < 15 ||
            agent->world().self().stamina() < 6000 )
        {
          doWait( agent );
          return true;
        }

        if( agent->world().self().unum() < 5 ||
            agent->world().self().unum() == 6  )
        {
           doMove( agent );
           return true;
        }


        if( agent->world().ball().pos().x < -10 &&
            ( agent->world().self().unum() > 8 ) )
        {
           doMove( agent );
           return true;
        }

        if( doMark( agent ) )
        {
           return true;
        }

        doMove( agent );

        return true;
}

void Bhv_MarkOpponent::doWait( PlayerAgent * agent )
{
      int S_time = agent->world().time().cycle();

      if( Body_GoToPoint( agent->world().ball().pos() , 1.0 , 15.0 ).execute( agent ) )
      {
          if( S_time - S_last_mark_time < 6 )
          {
            agent->setNeckAction( new Neck_ScanField() );
            return;
          }

          if( S_time - S_last_mark_time < 10 )
          {
            agent->setNeckAction( new Neck_TurnToBallOrScan() );
            return;
          }

          if( S_time - S_last_mark_time < 20 )
          {
            agent->setNeckAction( new Neck_ScanField() );
            return;
          }

         agent->setNeckAction( new Neck_ScanField() );
         return;
      }
      else
      {
         agent->setNeckAction( new rcsc::Neck_ScanField() );
         return;
      }
}

bool Bhv_MarkOpponent::doMark( PlayerAgent * agent )
{

      if( mark_opponent( agent ) && mark_opponent( agent ) != NULL &&
         !Circle2D( agent->world().ball().pos() , 8.0 ).contains( mark_opponent( agent )->pos() ) )
      {
        Vector2D extra_point( 0.0 , 0.0 );
        const WorldModel & wm = agent->world();
        Vector2D mark_point = mark_opponent( agent )->pos();

        double diff = mark_point.x - wm.ball().pos().x;

        if( wm.ball().pos().y > 28 )
        {
           if( diff > 7 )
           {
              extra_point.assign( -1.0 , 0.0 );
           }
           else if( diff < -7 )
           {
              extra_point.assign( 1.0 , 0.0 );
           }
           else
           {
              extra_point.assign( 0.0 , 1.0 );
           }
        }
        else if( wm.ball().pos().y < -28 )
        {
           if( diff > 7 )
           {
              extra_point.assign( -1.0 , 0.0 );
           }
           else if( diff < -7 )
           {
              extra_point.assign( 1.0 , 0.0 );
           }
           else
           {
              extra_point.assign( 0.0 ,-1.0 );
           }
        }
        else if( diff > 0 )
        {
           double diff_y = mark_point.y - wm.ball().pos().y ;
           if( diff_y > 5 )
           {
              extra_point.assign( -0.7 , -0.7 );
           }
           else if( diff_y < -5 )
           {
              extra_point.assign( -0.7 ,  0.7 );
           }
           else
           {
              extra_point.assign( -1.0 , 0.0 );
           }
        }
        else if( diff <= 0 )
        {
           double diff_y = mark_point.y - wm.ball().pos().y ;
           if( diff_y > 5 )
           {
              extra_point.assign( 0.7 , -0.7 );
           }
           else if( diff_y < -5 )
           {
              extra_point.assign( 0.7 ,  0.7 );
           }
           else
           {
              extra_point.assign( 1.0 , 0.0 );
           }
        }
        else
        {
           extra_point.assign( 0 , 0 );
        }

        double dist_thr = wm.self().pos().dist(mark_point+extra_point)*0.1;

        if( wm.self().pos().dist( mark_point + extra_point * 2 ) < 1 )
        {
           S_stop_move = false;
        }
        else
        {
           S_stop_move = true;
        }

        if( Body_GoToPoint2009( mark_point + extra_point * 2,
                                dist_thr,
                                ServerParam::i().maxPower() ).execute( agent ) &&
            S_stop_move )
        {
           agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        else
        {
           Body_TurnToPoint( mark_opponent( agent )->pos() ).execute( agent );
        }
      }
      else
      {
        Bhv_SetPlay( home_pos ).execute( agent );
      }

      return true;
}

void Bhv_MarkOpponent::doMove( PlayerAgent * agent )
{
       switch( agent->world().gameMode().type() )
       {
         case rcsc::GameMode::FreeKick_ :
            Bhv_SetPlayFreeKick( home_pos ).execute( agent );
            break;
         case rcsc::GameMode::KickIn_ :
            Bhv_SetPlayKickIn( home_pos ).execute( agent );
            break;
         case rcsc::GameMode::CornerKick_ :
            Bhv_SetPlayKickIn( home_pos ).execute( agent );
            break;
         case rcsc::GameMode::GoalKick_ :
            Bhv_SetPlayKickIn( home_pos ).execute( agent );
            break;
         default:
            Bhv_SetPlay( home_pos ).execute( agent );
            break;
       }
}

const PlayerObject * Bhv_MarkOpponent::
mark_opponent( PlayerAgent * agent )
{
        const WorldModel & wm = agent->world();
        WorldModelHighLevel WM;
        const PlayerPtrCont::const_iterator o_beg = wm.opponentsFromSelf().begin();
        const PlayerPtrCont::const_iterator o_end = wm.opponentsFromSelf().end  ();
        const PlayerObject * near_opp = wm.getOpponentNearestTo( home_pos , 10 , NULL );

        if( near_opp && wm.countTeammatesIn( Circle2D( near_opp->pos() , 1.5 ) , 1  , true ) == 0 )
        {
          return near_opp;
        }

        for( PlayerPtrCont::const_iterator o = o_beg ; o != o_end ; ++o )
        {
          if( WM.isBetween( (*o)->unum() , 0 , 12 ) && (*o)->distFromSelf() < 10.0 &&
               wm.countTeammatesIn( Circle2D( (*o)->pos() , 1.5 ) , 1  , true ) == 0 )
          {
             return (*o);
          }
        }

        const PlayerPtrCont::const_iterator o_beg_2 = wm.opponentsFromBall().begin();
        const PlayerPtrCont::const_iterator o_end_2 = wm.opponentsFromBall().end  ();

        for( PlayerPtrCont::const_iterator o = o_beg_2 ; o != o_end_2 ; ++o )
        {
          if( WM.isBetween( (*o)->unum() , 0 , 12 ) && (*o)->distFromSelf() < 10.0 &&
               wm.countTeammatesIn( Circle2D( (*o)->pos() , 1.5 ) , 1  , true ) == 0 )
          {
             return (*o);
          }
        }

        const PlayerObject * near_opp_2 = wm.getOpponentNearestTo( home_pos , 9 , NULL );
        if( near_opp_2 && wm.countTeammatesIn( Circle2D( near_opp_2->pos() , 1.5 ) , 1  , true ) == 0 )
        {
          return near_opp_2;
        }

        return NULL;
}
