// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "bhv_offensive_move.h"

bool Bhv_OffensiveMove::execute( rcsc::PlayerAgent * agent )
{
	// tackle
	
	if ( Bhv_BasicTackle( 0.9, 90.0 ).execute( agent ) )
	{
		return true;
	}

	const rcsc::WorldModel & wm = agent->world();
	WorldModelHighLevel WM;
	

	// check ball owner

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	
	if ( ! wm.existKickableTeammate()
		&& ( self_min <= 3
		|| ( self_min < mate_min + 3
			&& self_min < opp_min + 4 ) ) )
	{
		bool save_recovery = true;
		if( wm.ball().distFromSelf() < 5 )
			save_recovery = false;

		rcsc::Body_Intercept2008( save_recovery ).execute( agent );

		if ( wm.ball().distFromSelf()
		< rcsc::ServerParam::i().visibleDistance() )
		{
			agent->setNeckAction( new Neck_CheckBallOwner() );;
		}
		else
		{
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
		}
		return true;
	}
	
	if ( self_min <= mate_min + 1 && opp_min >= 2 )
	{
		bool save_recovery = true;
		if( wm.ball().distFromSelf() < 5 )
			save_recovery = false;

		Body_Intercept2008( save_recovery ).execute( agent );
	#if 0
		if ( self_min == 4 )
		{
		agent->setViewAction( new View_Wide() );
		}
		agent->setNeckAction( new Neck_TurnToBallOrScan() );
	#else
		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	#endif
		return true;
	}

	if( doForestall( agent ) );
	if( doIntercept( agent ) );

	const double dash_power = getDashPower( agent, M_home_pos );
	
	double dist_thr = wm.ball().distFromSelf() * 0.1;
	if ( dist_thr < 1.0 ) dist_thr = 1.0;

	agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
	agent->debugClient().setTarget( M_home_pos );

	bool save_stamina = true;
	if ( wm.ball().distFromSelf() < 20 )
		save_stamina = false;

	if ( wm.ball().distFromSelf() < 8 )
	{
		Bhv_GoToPointLookBall( M_home_pos,
					dist_thr,
					dash_power).execute( agent );
		return true;
	}
	else if ( ! rcsc::Body_GoToPoint( M_home_pos, dist_thr, dash_power
					).execute( agent ) )
	{
		rcsc::Body_TurnToBall().execute( agent );
	}
	
	if ( wm.existKickableOpponent()
	     && wm.ball().distFromSelf() < 18.0 )
	{
		agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	}
	else
	{
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	}
	
	return true;
}

double Bhv_OffensiveMove::getDashPower( const rcsc::PlayerAgent * agent,
                                        const rcsc::Vector2D & target_point )
{
	static bool s_recover_mode = false;
	
	const rcsc::WorldModel & wm = agent->world();
	/*--------------------------------------------------------*/
	// check recover
	
	if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
	{
	s_recover_mode = true;
	}
	else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
	{
	s_recover_mode = false;
	}
	
	/*--------------------------------------------------------*/
	double dash_power = rcsc::ServerParam::i().maxDashPower();
	const double my_inc
	= wm.self().playerType().staminaIncMax()
	* wm.self().recovery();
	
	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	
	if ( wm.defenseLineX() > wm.self().pos().x
		&& wm.ball().pos().x < wm.defenseLineX() + 20.0 )
	{
	// keep max power
		dash_power = rcsc::ServerParam::i().maxDashPower();
	}
	else if ( ! wm.existKickableTeammate()
		&& opp_min < self_min
		&& opp_min < mate_min
		&& target_point.x < wm.self().pos().x
		&& target_point.x < -25.0
		&& wm.ball().inertiaPoint( opp_min ).x < -25.0 )
	{
		dash_power = rcsc::ServerParam::i().maxDashPower();
	}
	// !exist teammate and opponent in round of ball
	else if ( self_min < mate_min && self_min < opp_min
		  &&!wm.existKickableOpponent() && !wm.existKickableTeammate()
		  && wm.ball().distFromSelf() < 3.0 )
	{
		dash_power = rcsc::ServerParam::i().maxPower();
	}
	else if ( s_recover_mode )
	{
		dash_power = my_inc - 25.0; // preffered recover value
		if ( dash_power < 0.0 ) dash_power = 0.0;
	}
	// exist kickable teammate
	else if ( wm.existKickableTeammate()
		&& wm.ball().distFromSelf() < 20.0 )
	{
		dash_power = std::min( my_inc * 1.1,
				       rcsc::ServerParam::i().maxDashPower() );
	}
	// in offside area
	else if ( wm.self().pos().x > wm.offsideLineX() )
	{
		dash_power = rcsc::ServerParam::i().maxDashPower();
	}
	// normal
	else
	{
		dash_power = std::min( my_inc * 1.7,
				       rcsc::ServerParam::i().maxDashPower() );
	}

	return dash_power;
}

bool Bhv_OffensiveMove::doForestall( rcsc::PlayerAgent * agent )
{
	const WorldModel & wm = agent->world();
	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	
	if ( wm.self().stamina() > ServerParam::i().staminaMax() * 0.6
		&& opp_min < 3
		&& opp_min < mate_min - 2
		&& ( opp_min < self_min - 2
		|| opp_min == 0
		|| ( opp_min == 1 && self_min > 1 ) )
		&& wm.ball().pos().dist( M_home_pos ) < 10.0
		&& wm.ball().distFromSelf() < 15.0 )
	{
		dlog.addText( Logger::TEAM,
			__FILE__": (doForestall) try. self=%d, mate=%d, opp=%d",
			self_min, mate_min, opp_min );
	
		if ( Body_ForestallBlock( M_home_pos ).execute( agent ) )
		{
		dlog.addText( Logger::TEAM,
				__FILE__": (doForestall) performed" );
		agent->debugClient().addMessage( "Attack:Forestall" );
		agent->setNeckAction( new Neck_TurnToBall() );
		return true;
		}
	}
	
	return false;
}

bool Bhv_OffensiveMove::doIntercept( rcsc::PlayerAgent * agent )
{
	const WorldModel & wm = agent->world();
	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	//int opp_min = wm.interceptTable()->opponentReachCycle();
	
	if ( ( mate_min >= 2
		&& self_min <= 4 )
		|| ( self_min <= mate_min + 1
		&& mate_min >= 4 )
		)
	{
		dlog.addText( Logger::TEAM,
			__FILE__": (doIntercept) performed" );
		agent->debugClient().addMessage( "Attack:Intercept" );
		Vector2D face_point( 52.5, wm.self().pos().y );
		if ( wm.self().pos().absY() < 10.0 )
		{
		face_point.y *= 0.8;
		}
		else if ( wm.self().pos().absY() < 20.0 )
		{
		face_point.y *= 0.9;
		}
	
		Body_Intercept2008( true , face_point ).execute( agent );
		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
		return true;
	}
	
	return false;
}
