// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef BHV_PRO_GOALIE_KICK_H
#define BHV_PRO_GOALIE_KICK_H

#include <rcsc/player/soccer_action.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_hold_ball2008.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/obsolete/body_kick_multi_step.h>

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/sector_2d.h>

#include "body_clear_ball2008.h"
#include "WorldModel.h"

static int S_counter_time = 0;

class Bhv_ProGoalieKick
    : public rcsc::SoccerBehavior {
private:

public:
    Bhv_ProGoalieKick()
      { }

    bool execute( rcsc::PlayerAgent * agent )
    {
	const rcsc::WorldModel & wm = agent->world();

	if( !wm.self().isKickable() )
	{
		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" ball is not kickable so return false." );
		return false;
	}

	int opp_min  = wm.interceptTable()->opponentReachCycle();
// 	int mate_min = wm.interceptTable()->teammateReachCycle();

	const rcsc::PlayerObject * near_opp =
	      wm.getOpponentNearestToSelf( 10 , false );

	if( S_counter_time >= 100 )
	{
		S_counter_time = 0;
		rcsc::dlog.addText( rcsc::Logger::ROLE,
					__FILE__" body clear ball to a best point now" );
		Body_ClearBall2008().execute( agent );
		agent->setNeckAction( new rcsc::Neck_ScanField() );
		return true;
	}

	if( near_opp && opp_min > 6 )
	{
		rcsc::Vector2D face_point( 100.0, 0.0 );
		rcsc::Body_HoldBall2008( true , face_point ).execute( agent );
		agent->setNeckAction( new rcsc::Neck_ScanField() );

		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" body hold ball -- opp_min > 3 " );
		S_counter_time ++;
		return true;
	}

	const rcsc::PlayerObject * near_mate =
	      wm.getTeammateNearestToSelf( 10 , false );

	if( near_mate && near_mate->pos().dist( wm.self().pos() ) > 15.0 )
	{
		rcsc::Vector2D face_point( 100.0, 0.0 );
		rcsc::Body_HoldBall2008( true , face_point ).execute( agent );
		agent->setNeckAction( new rcsc::Neck_ScanField() );

		rcsc::dlog.addText( rcsc::Logger::ROLE,
				    __FILE__" body hold ball -- near_mate far away" );
		S_counter_time ++;
		return true;
	}


	if( wm.ball().pos().absY() > 24.0 )
	{
		WorldModelHighLevel WM;
		rcsc::Vector2D * pos_shoot_ = new rcsc::Vector2D [2];
		pos_shoot_[0] = rcsc::Vector2D( wm.self().pos().x + 10 ,
					        34 * WM.sign(agent->world().ball().pos().y ) );

		pos_shoot_[1] = rcsc::Vector2D( wm.self().pos().x + 20 ,
					        34 * WM.sign(agent->world().ball().pos().y ) );

		int * nr = new int [2];

		for( int i = 0  ; i < 2 ; i ++ )
		{
			nr[i] = wm.countOpponentsIn( rcsc::Sector2D( wm.self().pos(),
								     0.5,
								     pos_shoot_[i].dist( wm.self().pos() ),
								     (pos_shoot_[i] - wm.self().pos()).th() - 20,
								     (pos_shoot_[i] - wm.self().pos()).th() + 20 ),
								     10,
								     false );
		}

		if( nr[0] > nr[1] && nr[1] == 0 )
			rcsc::Body_KickMultiStep(  pos_shoot_[1] , rcsc::ServerParam::i().maxPower() ).execute( agent );
		else if( nr[0] == 0 )
			rcsc::Body_KickMultiStep(  pos_shoot_[0] , rcsc::ServerParam::i().maxPower() ).execute( agent );
		else
			Body_ClearBall2008().execute(agent);

		agent->setNeckAction( new rcsc::Neck_ScanField() );
		delete [] nr;
		delete [] pos_shoot_;
		return true;
	}

	rcsc::dlog.addText( rcsc::Logger::ROLE,
				__FILE__" body clear ball to a best point now" );

	Body_ClearBall2008().execute( agent );
	agent->setNeckAction( new rcsc::Neck_ScanField() );
	return true;

    }

};

#endif
