
/*
 *Copyright:

 Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "body_penalty_area_pass.h"

bool Body_PenaltyAreaPass::execute( PlayerAgent * agent )
{

	const AbstractPlayerObject * teammates[11] =
	{
		agent->world().teammate( 1 ),
		agent->world().teammate( 2 ),
		agent->world().teammate( 3 ),
		agent->world().teammate( 4 ),
		agent->world().teammate( 5 ),
		agent->world().teammate( 6 ),
		agent->world().teammate( 7 ),
		agent->world().teammate( 8 ),
		agent->world().teammate( 9 ),
		agent->world().teammate(10 ),
		agent->world().teammate(11 )
	};

	Vector2D teammate[11] = {Vector2D( 1000 , 0 ) };

	/*-------------------------------------------------*/

	for( int i = 0 ; i < 11 ; i ++ )
	{
		if( teammates[i] )
		{
			teammate[i] = teammates[i]->pos();
		}
		else
			teammate[i].x = 1000;
	}

	/*-------------------------------------------------*/

	int point[11] = {0};
	for( int j = 0 ; j < 11 ; j ++ )
	{
		get_point_for_pos( agent , teammate[j] , point[j] );
	}


	/*-------------------------------------------------*/

	int i_max_point = 0;
	for( int l = 0 ; l < 11 ; l ++ )
		if( point[l] >= point[i_max_point] )
			i_max_point = l;

	/*-------------------------------------------------*/

	Body_KickMultiStep( teammate[i_max_point] ,
			    getEndSpeedForPass( agent , teammate[i_max_point] ) + 10 ).execute( agent );
	return true;

}


void Body_PenaltyAreaPass::get_point_for_pos( PlayerAgent * agent , Vector2D v , int & point )
{
	/*-------------------------------------------------*/
	if( v.x == 1000 ||
	    agent->world().self().pos().dist( v ) <= 0.2 )
	{
		point = -1000;
		return;
	}
	/*-------------------------------------------------*/
	if( prob( agent , v ) >= 0.8 )
		point = 5;
	else if( prob( agent , v ) >= 0.6 )
		point = 3;

	/*-------------------------------------------------*/

	return;
}

double Body_PenaltyAreaPass::prob( PlayerAgent * agent , Vector2D v )
{
	/*-------------------------------------------------*/
	Triangle2D tri_me_to_pos = WM.Create_Triangle_From_Two_Points( agent->world().self().pos(), v, 6  );
	Triangle2D tri_pos_to_goal ( v ,
				     Vector2D( 52.5 ,  7.0 ),
				     Vector2D( 52.5 , -7.0 ) );
	/*-------------------------------------------------*/
	double probablity = 0.0;

	const PlayerObject * opp_goalie = agent->world().getOpponentGoalie();
	Vector2D       opp_goalie_pos( 1000 , 0 );
	if( opp_goalie ) opp_goalie_pos = opp_goalie->pos();
	/*-------------------------------------------------*/

	if( !WM.isInTheirPenaltyArea( agent , v  ) )
		return 0.0;

	if( WM.getNrInSetInTriangle( agent , tri_me_to_pos   ) == 0 )
		probablity += 0.2;

	if( !opp_goalie )
		probablity += 0.2;
	else if( !tri_me_to_pos.contains( opp_goalie_pos ) )
		probablity += 0.2;
	else
		probablity -= 0.2;

	/*-------------------------------------------------*/

	if( WM.getNrInSetInTriangle( agent , tri_pos_to_goal ) == 0 )
		probablity += 0.2;


	if( !opp_goalie )
		probablity += 0.2;
	else if( !tri_pos_to_goal.contains( opp_goalie_pos ) )
		probablity += 0.2;
	else
		probablity -= 0.2;

	/*-------------------------------------------------*/

	double distance_pos_to_goal = v.dist( Vector2D( 52.5 , 0.0 ) );
	double distance_opg_to_goal = -1000;
	if( opp_goalie ) distance_opg_to_goal = opp_goalie_pos.dist( Vector2D( 52.5 , 0.0 ) );


	if( !opp_goalie )
		probablity += 0.1;
	else if( distance_opg_to_goal > distance_pos_to_goal )
		probablity += 0.1;


	/*-------------------------------------------------*/

	Circle2D around_circle( v , 5.0 );


	if( !opp_goalie )
		probablity += 0.1;
	else if( !around_circle.contains( opp_goalie_pos ) )
		probablity += 0.1;
	else
		probablity -= 0.2;

	/*-------------------------------------------------*/

	return probablity;
}


double Body_PenaltyAreaPass::getEndSpeedForPass( rcsc::PlayerAgent * agent , rcsc::Vector2D posPass )
{
		rcsc::Vector2D posAgent = agent->world().self().pos();

		int dSpeed = 0;

		if ( posAgent.dist(posPass) > 50 )
			dSpeed = 32;
		else if ( posAgent.dist(posPass) > 40 )
			dSpeed = 28;
		else if ( posAgent.dist(posPass) > 30 )
			dSpeed = 18;
		else if ( posAgent.dist(posPass) > 25 )
			dSpeed = 12;
		else if ( posAgent.dist(posPass) > 20 )
			dSpeed = 10;
		else
			dSpeed = 7;
	
		return dSpeed;
}
