// -*-c++-*-


/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA, Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "neck_default_intercept_neck.h"

#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/view_wide.h>
#include <rcsc/action/view_normal.h>
#include <rcsc/action/view_synch.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

/*-------------------------------------------------------------------*/
/*!

*/

bool Neck_DefaultInterceptNeck::execute( rcsc::PlayerAgent * agent )
{
	
	if ( doTurnToReceiver( agent ) )
	{
		return true;
	}
	
	rcsc::dlog.addText( rcsc::Logger::TEAM,
				__FILE__": Neck_DefaultInterceptNeck. default action" );
	
	if ( M_default_view_act )
	{
		M_default_view_act->execute( agent );
		agent->debugClient().addMessage( "InterceptNeck:DefaultView" );
	}

	if ( M_default_neck_act )
	{
		M_default_neck_act->execute( agent );
		agent->debugClient().addMessage( "InterceptNeck:DefaultNeck" );
	}
	else
	{
		rcsc::Neck_TurnToBallOrScan().execute( agent );
		agent->debugClient().addMessage( "InterceptNeck:BallOrScan" );
	}
	return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool Neck_DefaultInterceptNeck::doTurnToReceiver( rcsc::PlayerAgent * agent )
{
	rcsc::dlog.addText( rcsc::Logger::TEAM,
				__FILE__": doTurnToReceiver()" );
	const rcsc::WorldModel & wm = agent->world();
	
	if ( wm.interceptTable()->selfReachCycle() >= 2 )
	{
		rcsc::dlog.addText( rcsc::Logger::TEAM,
				__FILE__": doTurnToReceiver() self reach cycle >= 2 " );
		return false;
	}
	
// 	const rcsc::Body_Pass::PassRoute * pass/* = rcsc::Bhv_PassTest::get_best_pass( wm )*/;
// 	if ( ! pass )
// 	{
// 		rcsc::dlog.addText( rcsc::Logger::TEAM,
// 				__FILE__": doTurnToReceiver() no pass." );
// 		return false;
// 	}
// 	
	rcsc::Vector2D target;
	double   first_s;
	int      receiver_n;
	if( !rcsc::Body_Pass::get_best_pass( agent->world() , &target , &first_s , &receiver_n ) )
	{
	
		const rcsc::AbstractPlayerObject * receiver_ = agent->world().teammate( receiver_n );

		if ( !receiver_ )
			return false;

		if ( receiver_->posCount() <= 0 )
		{
			rcsc::dlog.addText( rcsc::Logger::TEAM,
					__FILE__": doTurnToReceiver() receiver pos_count %d <= 1",
					receiver_->posCount() );
			return false;
		}
		
		rcsc::View_Synch().execute( agent );
		
		rcsc::Vector2D face_point = receiver_->pos();
		if ( receiver_->distFromBall() > 5.0
			&& receiver_->pos().x < 35.0 )
		{
			face_point.x += 3.0;
		}
		
		const rcsc::Vector2D next_pos = agent->effector().queuedNextMyPos();
		const rcsc::AngleDeg next_body = agent->effector().queuedNextMyBody();
		const double next_view_half_width = agent->effector().queuedNextViewWidth().width() * 0.5;
		
		rcsc::AngleDeg rel_angle = ( face_point - next_pos ).th() - next_body;
		
		if ( rel_angle.abs() > rcsc::ServerParam::i().maxNeckAngle() + next_view_half_width - 5.0 )
		{
			rcsc::dlog.addText( rcsc::Logger::TEAM,
					__FILE__": Neck_DefaultInterceptNeck.cannot face to receiver %d (%.1f %.1f)",
					receiver_->unum(),
					face_point.x, face_point.y );
		
			return false;
		}
		
		rcsc::dlog.addText( rcsc::Logger::TEAM,
					__FILE__": Neck_DefaultInterceptNeck.(%.1f %.1f)",
					face_point.x, face_point.y );
		rcsc::Neck_TurnToPoint( face_point ).execute( agent );
		
		agent->debugClient().setTarget( receiver_->unum() );
		agent->debugClient().addMessage( "InterceptNeck:toReceiver" );
		agent->debugClient().addLine( wm.self().pos(), face_point );
		agent->debugClient().addCircle( face_point, 3.0 );
		return true;
	}
	else
		return false;
}

/*-------------------------------------------------------------------*/
/*!

*/

rcsc::NeckAction * Neck_DefaultInterceptNeck::clone() const
{
	return new Neck_DefaultInterceptNeck( ( M_default_view_act
						? M_default_view_act->clone()
						: static_cast< rcsc::ViewAction * >( 0 ) ),
						( M_default_neck_act
						? M_default_neck_act->clone()
						: static_cast< rcsc::NeckAction * >( 0 ) ) );
}
