// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA
 Modified Copyright (C) Hossein Mobasher


 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_center_forward.h"

#include "strategy.h"


#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"


#include <iostream>
#include <math.h>

#include <rcsc/geom/sector_2d.h>
#include <rcsc/geom/line_2d.h>


#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#include <rcsc/action/obsolete/body_kick_multi_step.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/bhv_shoot2008.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_hold_ball2008.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_hold_ball2008.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_intercept2008.h>

#include "bhv_penalty_kick.h"
#include "WorldModel.h"
#include "bhv_dribble.h"
#include "bhv_dribble2009.h"
#include "body_penalty_area_pass.h"
#include "best_shoot_point.h"
#include "body_kick_to_corner.h"
#include "bhv_cross_to_center.h"
#include "body_go_to_point_2009.h"
#include "bhv_attacking_move.h"
#include "bhv_self_pass.h"
#include "bhv_cross.h"
#include "bhv_shoot2009.h"

#include "bhv_basic_tackle.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::execute( rcsc::PlayerAgent * agent )
{

	bool kickable = agent->world().self().isKickable();

	if ( agent->world().existKickableTeammate()
		&& agent->world().teammatesFromBall().front()->distFromBall()
		< agent->world().ball().distFromSelf() )
	{
		kickable = false;
	}


	if( kickable )
	{

		doKick( agent );
	}
	else
	{

		if ( !agent->world().existKickableTeammate() && !agent->world().existKickableOpponent() &&
		      agent->world().ball().pos().x > 10 &&
		      agent->world().countTeammatesIn( Circle2D( agent->world().ball().pos() , 10 ) , 5 , false ) == 0 )
		{
			Body_Intercept2008().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			return;
		}

		doMove( agent );
	}

}

/*------------------------------------------------------------------*/
/*!

*/

void RoleCenterForward::doKick( rcsc::PlayerAgent * agent )
{

    WorldModelHighLevel WM;
    const WorldModel & wm = agent->world();
    Vector2D my_pos = wm.self().pos();

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    {
	rcsc::Body_Pass().execute( agent );
	break;
    }
    case Strategy::BA_Stopper:
    {
	bool left = ( agent->world().self().pos().y < 0.0 ? true : false );

	if( agent->world().self().pos().x < -26.25 )
	{
		if ( rcsc::Body_Pass().execute( agent ) )
			;
		else
			Body_KickToCorner( left ).execute( agent );
	}
	else
	{
		rcsc::Circle2D   round_c ( my_pos , 5.0 );
		rcsc::Triangle2D front_t ( my_pos - Vector2D( 5.0  ,  0.0 ),
					   my_pos + Vector2D( 15.0 , -8.0 ),
					   my_pos + Vector2D( 15.0 ,  8.0 ));

		if( wm.countOpponentsIn( front_t, 15, false ) == 0 && wm.countOpponentsIn( round_c , 5 , false ) == 0 )
		{
			if ( Bhv_Dribble().execute(agent) )
				;
			else
			{
				Bhv_Dribble2009().execute( agent );
			}
		}
		else
			if ( rcsc::Body_Pass().execute( agent ) )
				;
			else
				Body_KickToCorner( left ).execute( agent );
	}

	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	break;
    }
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    {
	doOffensiveMiddleKick( agent );
	break;
    }
    case Strategy::BA_Cross:
    {
	Bhv_BasicOffensiveKick().execute( agent );
	break;
    }
    case Strategy::BA_ShootChance:
    {
	if( Bhv_Shoot2008().execute( agent ) )
	{
		break;
	}
	if( Bhv_Shoot2009().execute( agent ) )
	{
		break;
	}

	Vector2D posAgent = wm.self().pos();
	AngleDeg ang_rel_to_goal = ( Vector2D( 50.0 , 0.0 ) - posAgent ).th();
	double   dist_to_goal    = posAgent.dist( Vector2D( 50.0 , 0.0 ) );
	Sector2D sector_to_goal ( posAgent,
				  0.5,
				  dist_to_goal - 1.0,
				  ang_rel_to_goal - 15,
				  ang_rel_to_goal + 15
				);

	Circle2D circle_around_self( posAgent , 2.0 );

	if( wm.countOpponentsIn( sector_to_goal     , 10 , false ) == 0 &&
	    wm.countOpponentsIn( circle_around_self , 10 , false ) == 0 )
	{
		int dash_step = 10;

		const rcsc::Rect2D target_rect
			  = rcsc::Rect2D::from_center( wm.ball().pos().x + 5.5,
						       wm.ball().pos().y,
						       10.0, 12.0 );

	
		if ( wm.existOpponentIn( target_rect, 10, false ) )
		{
			dash_step = 3;
		}

		Body_Dribble2008( Vector2D( 50.0 , 0.0 ),
				  1.0,
				  rcsc::ServerParam::i().maxPower(),
				  dash_step,
				  false ).execute( agent );
		agent->setNeckAction( new rcsc::Neck_ScanField() );
		return;
	}

    }
    default:
    {
	if( Circle2D(Vector2D(52.5,0.0),18.0).contains(wm.self().pos()) )
	{
		const Vector2D posAgent = wm.self().pos();
	
		const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
		
		rcsc::Vector2D opp_goalie_pos  ( -1000.0, 0.0 );

		if ( opp_goalie  ) opp_goalie_pos   = opp_goalie->pos()  ;
		
		
		rcsc::Triangle2D triangle_goal1( wm.self().pos() ,
						rcsc::Vector2D(52.5, 0.0),
						rcsc::Vector2D(52.5, 8.0));
		
		rcsc::Triangle2D triangle_goal2( wm.self().pos() ,
						rcsc::Vector2D(52.5,-8.0),
						rcsc::Vector2D(52.5, 0.0));

		Vector2D posMirror( 52.5 , posAgent.y );
		
		if( Bhv_Shoot2008().execute( agent ) )
		{
			break;
		}
		else if( opp_goalie && fabs(posMirror.y) <= 10.0 && !triangle_goal1.contains( opp_goalie_pos ) )
		{
			rcsc::Body_KickMultiStep( Vector2D( 52.5 , 6.5 ) , 100 ).execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else if( opp_goalie && fabs(posMirror.y) <= 10.0 && !triangle_goal2.contains( opp_goalie_pos ) )
		{
			rcsc::Body_KickMultiStep( Vector2D( 52.5 ,-6.5 ) , 100 ).execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else
		{
			Body_PenaltyAreaPass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
	}

	Bhv_BasicOffensiveKick().execute( agent );
	break;
    }
    }
}
/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterForward::doMove( rcsc::PlayerAgent * agent )
{

    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
          && ! wm.existKickableOpponent() )
     {
         ball_reach_step
             = std::min( wm.interceptTable()->teammateReachCycle(),
                         wm.interceptTable()->opponentReachCycle() );
     }
     const rcsc::Vector2D base_pos
         = wm.ball().inertiaPoint( ball_reach_step );
 
 	rcsc::Vector2D ball_pos
 	= wm.ball().pos();
 
     rcsc::Vector2D home_pos
         = formation().getPosition( agent->config().playerNumber(),
                                    base_pos );
     if ( rcsc::ServerParam::i().useOffside() )
     {
         home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
     }
	
	if( home_pos.dist( agent->world().ball().pos() ) < 4.0 )
		home_pos = ball_pos;

	switch ( Strategy::get_ball_area( base_pos ) ) {
	case Strategy::BA_CrossBlock:
	case Strategy::BA_Stopper:
	case Strategy::BA_Danger:
	case Strategy::BA_DribbleBlock:
	case Strategy::BA_DefMidField:
	   Bhv_AttackingMove( home_pos , true ).execute( agent );
	   break;
	case Strategy::BA_DribbleAttack:
	case Strategy::BA_OffMidField:
	   Bhv_AttackingMove( home_pos , true ).execute( agent );
	   break;
	case Strategy::BA_Cross:
	case Strategy::BA_ShootChance:
	{
// 		doShootAreaMove( agent , home_pos );
		Bhv_AttackingMove( home_pos , true ).execute( agent );
	   	break;

	}
	default:
		Bhv_BasicMove( home_pos ).execute( agent );
		break;
	}
}

/*---------------------------------------------------*/
/*!

*/
bool
RoleCenterForward::isOppGoalieInRect( rcsc::PlayerAgent * agent , int iSide )
{
   rcsc::Rect2D rect;
   if( iSide == 1 )
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9,-9.12) , rcsc::Vector2D(52.5,0.0 ) );//right
   else if( iSide == 2 )
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9, 9.12) , rcsc::Vector2D(52.5,0.0 ) );//left
   else
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9,-9.12) , rcsc::Vector2D(52.5,9.12) );//left and right

    const rcsc::PlayerObject * opp_goalie = agent->world().getOpponentGoalie();

    if( !opp_goalie )
       return false;
    else
    {
	if( rect.contains(opp_goalie->pos()))
		return true;
	else
		return false;
     }
}

/*----------------------------------------------*/
/*!

*/
void
RoleCenterForward::doCrossAreaKick( rcsc::PlayerAgent* agent )
{
    const rcsc::WorldModel& wm = agent->world();

    //-------------------------------------------------------
    // !!!check stamina!!!
    static bool S_recover_mode = false;
    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.6 )
    {
        S_recover_mode = false;
    }

    //-------------------------------------------------------
    // check opponent fielder & goalie distance
    const double opp_dist
        = ( wm.opponentsFromSelf().empty()
            ? 1000.0
            : wm.opponentsFromSelf().front()->distFromSelf() );

    const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
    bool goalie_near = false;
    bool goalie_verynear = false;
    if ( opp_goalie )
    {
        if ( opp_goalie->pos().x > 32.0
             && opp_goalie->pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() )
        {
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 4.0 ) )
            {
                goalie_near = true;
            }
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 2.0 ) )
            {
                goalie_verynear = true;
            }
        }
    }

    //-------------------------------------------------------
    if ( goalie_verynear )
    {
        // kick to the near side corner
        agent->debugClient().addMessage( "XAreaGKNear" );

	Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------

    WorldModelHighLevel WorldModel;
    rcsc::Vector2D pass_point;
    int receiver_unum = rcsc::Unum_Unknown;
    const bool can_pass = rcsc::Body_Pass::get_best_pass(  wm, &pass_point,
                                                              NULL, &receiver_unum );
    const rcsc::PlayerObject * receiver = ( receiver_unum != rcsc::Unum_Unknown
                                            ? WorldModel.getTeammate( receiver_unum , agent )
                                            : NULL );
    bool receiver_free
        = ( can_pass
            && receiver
            &&  ! wm.existOpponentIn( rcsc::Circle2D( receiver->pos(), 5.0 ), 10, true ) );

    if ( can_pass
         && pass_point.x > 37.0
         && receiver_free
         && pass_point.absY() < 12.0
         && ! wm.existOpponentIn( rcsc::Circle2D( pass_point, 6.0 ), 10, true )
         && ( ! opp_goalie
              || opp_goalie->pos().dist( pass_point ) > 6.0 )
         )
    {
        agent->debugClient().addMessage( "XAreaPassChance1" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    if ( can_pass )
    {
        rcsc::Rect2D shoot_rect = rcsc::Rect2D::from_center( 52.5 - 7.0,
                                                             0.0,
                                                             14.0,
                                                             32.0 );
        if ( shoot_rect.contains( pass_point )
             && receiver_free
             && wm.countOpponentsIn( shoot_rect, 10, false ) <= 2
             && ! wm.existOpponentIn( rcsc::Circle2D( pass_point, 5.0 ), 10, true ) )
        {
            agent->debugClient().addMessage( "XAreaPassChance2" );
            rcsc::Body_Pass().execute( agent );

            if ( agent->effector().queuedNextBallKickable() )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            return;
        }
    }

    if ( can_pass
         && pass_point.x > 39.0
         && pass_point.absY() < 7.0 )
    {
        agent->debugClient().addMessage( "XAreaPass1" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    // very side attack
    if ( can_pass
         && wm.self().pos().absY() > 27.0
         && opp_dist < 4.0
         && pass_point.x > 26.0 )
    {
        agent->debugClient().addMessage( "XAreaPass2" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    //-------------------------------------------------------
    // cross to center
    // exist teammate at cross area
    if ( opp_dist < 3.0
         && Bhv_CrossToCenter::get_best_point( agent, NULL ) )
    {
        agent->debugClient().addMessage( "XAreaCross1" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

#if 1
    //-------------------------------------------------------
    // dribble cource is blocked
    if ( can_pass
         && receiver_free
         && wm.self().pos().x > 40.0
         && wm.self().pos().absY() > 20.0 )
    {
        rcsc::Vector2D rect_center( 47.0,
                                    ( wm.self().pos().absY() + 9.0 ) * 0.5 );
        if ( wm.self().pos().y < 0.0 ) rect_center.y *= -1.0;

        rcsc::Rect2D side_rect
            = rcsc::Rect2D::from_center( rect_center,
                                         10.0,
                                         wm.self().pos().absY() - 9.0 );
        agent->debugClient().addRectangle( side_rect );
        if ( wm.countOpponentsIn( side_rect, 10, false ) >= 3 ) // 2 )
        {
            agent->debugClient().addMessage( "XAreaPassBlocked" );
            rcsc::Body_Pass().execute( agent );

            if ( agent->effector().queuedNextBallKickable() )
            {
                if ( receiver->posCount() > 0 )
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToPlayerOrScan( receiver ) );
                }
                else
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                }
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            return;
        }
    }
#endif

    //-------------------------------------------------------
    if ( S_recover_mode
         && ! goalie_near
         && opp_dist > 4.0 )
    {
        agent->debugClient().addMessage( "XAreaHold" );
        rcsc::Body_HoldBall2008().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------
    // keep dribble
    if ( ! goalie_near
         && opp_dist > rcsc::ServerParam::i().tackleDist() + 0.5 )
    {
        agent->debugClient().addMessage( "XAreaDrib1" );
	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
        return;
    }
    //-------------------------------------------------------
    // forcely pass
    if ( can_pass )
    {
        agent->debugClient().addMessage( "XAreaPass3" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    //-------------------------------------------------------
    // not exist opp on my body dir
    rcsc::Sector2D body_sector( wm.self().pos(),
                                0.5, 8.0,
                                wm.self().body() - 25.0,
                                wm.self().body() + 25.0 );
    // oponent check with goalie
    if ( ! wm.existOpponentIn( body_sector, 10, true ) )
    {
        agent->debugClient().addMessage( "XAreaDrib2" );
	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
        return;
    }

    //-------------------------------------------------------
    // forcely cross
    if ( S_recover_mode && opp_dist < 2.5 )
    {
        agent->debugClient().addMessage( "XAreaCross2" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
}

/*----------------------------------------------*/
/*!

*/

void
RoleCenterForward::doDribbleAttackAreaKick( rcsc::PlayerAgent* agent )
{
    // !!!check stamina!!!
    static bool S_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        S_recover_mode = false;
    }

    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    rcsc::Vector2D nearest_opp_pos( -1000.0, 0.0 );
    if ( nearest_opp ) nearest_opp_pos = nearest_opp->pos();

#if 0
    for ( int dash_step = 16; dash_step >= 6; dash_step -= 2 )
    {
        if ( doSelfPass( agent, dash_step ) )
        {
            return;
        }
    }
#else
    if ( Bhv_SelfPass().execute( agent ) )
    {
        return;
    }
#endif

    if ( Bhv_Dribble().doLongDribble( agent ) )
    {
        return;
    }

    // recover mode
    if ( S_recover_mode )
    {
        if ( nearest_opp_dist > 5.0 )
        {
            rcsc::Body_HoldBall2008( true, rcsc::Vector2D(36.0, wm.self().pos().y * 0.9 )//0.0)
                                     ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
        else if ( rcsc::Body_Pass().execute( agent ) )
        {
            return;
        }
        else
        {
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
	if ( Bhv_CrossToCenter().execute( agent ) )
	        return;
    }

    if ( nearest_opp_dist < rcsc::ServerParam::i().defaultPlayerSize() * 2.0 + 0.2
         || wm.existKickableOpponent() )
    {
	Vector2D target_pass ( 0.0 , 0.0 );
	if ( Body_Pass().get_best_pass( wm , &target_pass , NULL , NULL ) )
	{
		if( target_pass.x > wm.self().pos().x )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
			return;
		}
		else
		{
			Body_ClearBall().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
			return;
		}
	}

        if ( wm.self().pos().x < 25.0 )
        {
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
        }
        else if ( wm.self().pos().x < 35.0 )
        {
            Bhv_CrossToCenter().execute( agent );
        }
        else
        {
            Bhv_Cross().execute( agent );
            return;
        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }


    if ( nearest_opp
         && wm.self().pos().x > 20.0
         && nearest_opp->distFromSelf() < 1.0
         && nearest_opp->pos().x > wm.self().pos().x
         && std::fabs( nearest_opp->pos().y - wm.self().pos().y ) < 0.9 )
    {
        Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                           ).execute( agent );

        agent->setNeckAction( new rcsc::Neck_ScanField() );
        agent->debugClient().addMessage( "SF:BaseKickAvoidOpp" );
        return;
    }

    // wing dribble
    doCrossAreaKick(agent);
}

//-----------------------------------------

void RoleCenterForward::doOffensiveMiddleKick( PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 6 );

    if ( nearest_opp )
    {

    }

    if ( Bhv_SelfPass().execute( agent ) )
    {
        return;
    }

    Vector2D receiver;

    bool pass = Body_Pass().get_best_pass( wm , &receiver , NULL , NULL );

    // dribble straight
    if ( wm.self().stamina() > //rcsc::ServerParam::i().staminaMax() * 0.5
         rcsc::ServerParam::i().recoverDecThrValue() + 200.0
         && ( ! nearest_opp
              || nearest_opp->pos().x < wm.self().pos().x + 0.5
              || std::fabs( nearest_opp->pos().y -  wm.self().pos().y ) > 2.0
              || nearest_opp->distFromSelf() > 2.0 )
         )
    {
        const rcsc::Rect2D target_rect
            = rcsc::Rect2D::from_center( wm.ball().pos().x + 5.5,
                                         wm.ball().pos().y,
                                         10.0, 12.0 );
        if ( ! wm.existOpponentIn( target_rect, 10, false ) )
        {
            rcsc::Vector2D drib_target( 47.0, wm.self().pos().y );
            if ( wm.self().body().abs() < 20.0 )
            {
                drib_target
                    = wm.self().pos()
                    + rcsc::Vector2D::polar2vector( 10.0, wm.self().body() );
            }

            int dash_step = 10;

            const rcsc::Rect2D safety_rect
                = rcsc::Rect2D::from_center( wm.ball().pos().x + 7.0,
                                             wm.ball().pos().y,
                                             14.0, 15.0 );
            if ( wm.existOpponentIn( target_rect, 10, false ) )
            {
                dash_step = 3;
            }

            agent->debugClient().addMessage( "MidDribSt" );
            agent->debugClient().addRectangle( target_rect );
            agent->debugClient().addRectangle( safety_rect );

            rcsc::Body_Dribble2008( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    dash_step, // dash count
                                    false // no dodge
                                    ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            return;
        }
    }


    // pass in order to avoid opponent
    if ( pass
         && nearest_opp
         && nearest_opp->distFromSelf() < 5.0 )
    {
        rcsc::Body_Pass().execute( agent );
        agent->debugClient().addMessage( "MidPass(1)" );
        return;
    }

    //------------------------------------------------
    // pass to side area
    rcsc::Rect2D front_rect( rcsc::Vector2D( wm.self().pos().x,
                                             wm.self().pos().y - 5.0 ),
                             rcsc::Size2D( 10.0, 10.0 ) );
    if ( pass
         && receiver.x > wm.self().pos().x
         && receiver.absY() > 20.0
         && wm.existOpponentIn( front_rect, 10, false ) ) // without goalie
    {
        agent->debugClient().addMessage( "MidPassSide(1)" );
        rcsc::Body_Pass().execute( agent );
        return;
    }

    if ( pass
         && receiver.x > wm.self().pos().x + 10.0
         && receiver.absY() > 20.0 )
    {
        agent->debugClient().addMessage( "MidPassSide(2)" );
        rcsc::Body_Pass().execute( agent );
        return;
    }

    rcsc::Vector2D drib_target( 50.0, 0.0 );
    drib_target.y = ( agent->world().self().pos().y > 0.0
                      ? 5.0 : -5.0 );

    //------------------------------------------------
    // opp's is behind of me
    if ( ! nearest_opp
         || ( nearest_opp->distFromSelf()
              > rcsc::ServerParam::i().defaultPlayerSize() * 2.0 + 0.2
              && nearest_opp->pos().x < wm.self().pos().x )
         )
    {
        rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();
        const rcsc::Sector2D sector( agent->world().self().pos(),
                                     0.5, 15.0,
                                     drib_angle - 30.0, drib_angle + 30.0 );
        // oppnent check with goalie
        if ( ! agent->world().existOpponentIn( sector, 10, true ) )
        {
            agent->debugClient().addMessage( "MidDribFast" );
            rcsc::Body_Dribble2008( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    5
                                    ).execute( agent );
        }
        else
        {
            agent->debugClient().addMessage( "MidDribSlow" );
            rcsc::Body_Dribble2008( drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower() * 0.5,
                                    5
                                    ).execute( agent );
        }
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // check pass point
    if ( pass
         && receiver.x > wm.offsideLineX() )
    {
        agent->debugClient().addMessage( "MidPassOverOffside" );
        rcsc::Body_Pass().execute( agent );
        return;
    }

    //------------------------------------------------
    // opp is far from me
    if ( ! nearest_opp
         || nearest_opp->distFromSelf() > 3.0 )
    {
        agent->debugClient().addMessage( "MidDribD1" );
        rcsc::Body_Dribble2008( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                1
                                ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }

    //------------------------------------------------
    // opp is close
    // pass
    if ( pass )
    {
        agent->debugClient().addMessage( "MidPassNorm" );
        rcsc::Body_Pass().execute( agent );
        return;
    }

    //------------------------------------------------
    // other
    // near opponent may exist.

    agent->debugClient().addMessage( "MidAreaHold" );
    rcsc::Vector2D face_point( 52.5, 0.0 );
    rcsc::Body_HoldBall2008( true, face_point ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );

}


void
RoleCenterForward::doShootAreaMove( rcsc::PlayerAgent* agent,
                                  const rcsc::Vector2D& home_pos )
{
    const rcsc::WorldModel & wm = agent->world();

    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    //------------------------------------------------------

    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ( ! wm.existKickableTeammate() && self_min < 4 )
         || self_min < mate_min
         || ( self_min <= mate_min && self_min > 5 )
         )
    {
        agent->debugClient().addMessage( "I am getting to Ball now" );

        rcsc::Body_Intercept2008().execute( agent );
#if 0
        if ( self_min == 4 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Wide() );
        }
        else if ( self_min == 3 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Normal() );
        }

        if ( wm.self().pos().x > 30.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
#else
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
#endif
        return;
    }

    //------------------------------------------------------
    if ( self_min < 20
         && self_min < mate_min
         && self_min < opp_min + 20
         && wm.ball().pos().absY() > 19.0
         //&& wm.interceptTable()->isOurTeamBallPossessor() )
         )
    {

        rcsc::Body_Intercept2008().execute( agent );
#if 0
        if ( self_min == 4 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Wide() );
        }
        else if ( self_min == 3 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Normal() );
        }

        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
#else
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
#endif
        return;
    }

    //------------------------------------------------------
    // decide move target point & dash power
    const rcsc::Vector2D target_point = getShootAreaMoveTarget( agent, home_pos );
    const double dash_power = getShootAreaMoveDashPower( agent, target_point );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doShootAreaMove. go to cross point (%.2f, %.2f)",
                        target_point.x, target_point.y );

    agent->debugClient().addMessage( "ShootMoveP%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );


    if ( wm.self().pos().x > target_point.x + dist_thr*0.5
         && std::fabs( wm.self().pos().x - target_point.x ) < 3.0
         && wm.self().body().abs() < 10.0 )
    {
        agent->debugClient().addMessage( "Back" );
        double back_dash_power
            = wm.self().getSafetyDashPower( -dash_power );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Back Move" );
        agent->doDash( back_dash_power );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power,
                                 5, // cycle
                                 false, // no back dash
                                 true, // save recovery
                                 20.0 // dir thr
                                 ).execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. already target point. turn to front" );
        rcsc::Body_TurnToAngle( 0.0 ).execute( agent );
    }

    if ( wm.self().pos().x > 35.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Neck to Goalie or Scan" );
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Neck to Ball or Scan" );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

rcsc::Vector2D
RoleCenterForward::getShootAreaMoveTarget( rcsc::PlayerAgent * agent,
                                         const rcsc::Vector2D & home_pos )
{

    const rcsc::WorldModel & wm = agent->world();

    rcsc::Vector2D target_point = home_pos;

    int mate_min = wm.interceptTable()->teammateReachCycle();
    rcsc::Vector2D ball_pos = ( mate_min < 10
                                ? wm.ball().inertiaPoint( mate_min )
                                : wm.ball().pos() );
    bool opposite_side = false;
    const PositionType position_type = Strategy::i().getPositionType( wm.self().unum() );

    if ( ( ball_pos.y > 0.0
           && position_type == Position_Left )
         || ( ball_pos.y <= 0.0
              && position_type == Position_Right )
         || ( ball_pos.y * home_pos.y < 0.0
              && position_type == Position_Center )
         )
    {
        opposite_side = true;
    }

    if ( opposite_side
         && ball_pos.x > 40.0 // very chance
         && 6.0 < ball_pos.absY()
         && ball_pos.absY() < 15.0 )
    {
        rcsc::Circle2D goal_area( rcsc::Vector2D( 47.0, 0.0 ),
                                  7.0 );

        if ( ! wm.existTeammateIn( goal_area, 10, true ) )
        {
            agent->debugClient().addMessage( "GoToCenterCross" );

            target_point.x = 47.0;

            if ( wm.self().pos().x > wm.offsideLineX() - 0.5
                 && target_point.x > wm.offsideLineX() - 0.5 )
            {
                target_point.x = wm.offsideLineX() - 0.5;
            }

            if ( ball_pos.absY() < 9.0 )
            {
                target_point.y = 0.0;
            }
            else
            {
                target_point.y = ( ball_pos.y > 0.0
                                   ? 1.0 : -1.0 );
                                   //? 2.4 : -2.4 );
            }

            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. center cross point" );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. exist teammate in cross point" );
        }
    }

    // consider near opponent
    {
        double opp_dist = 200.0;
        const rcsc::PlayerObject * nearest_opp
            = wm.getOpponentNearestTo( target_point, 10, &opp_dist );

        if ( nearest_opp
             && opp_dist < 2.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. Point Blocked. old target=(%.2f, %.2f)",
                                target_point.x, target_point.y );
            agent->debugClient().addMessage( "AvoidOpp" );
            //if ( nearest_opp->pos().x + 3.0 < wm.offsideLineX() - 1.0 )
            if ( nearest_opp->pos().x + 2.0 < wm.offsideLineX() - 1.0  )
            {
                //target_point.x = nearest_opp->pos().x + 3.0;
                target_point.x = nearest_opp->pos().x + 2.0;
            }
            else
            {
                target_point.x = nearest_opp->pos().x - 2.0;
#if 1
                if ( std::fabs( wm.self().pos().y - target_point.y ) < 1.0 )
                {
                    target_point.x = nearest_opp->pos().x - 3.0;
                }
#endif
            }
        }

    }

    // consider goalie
    if ( target_point.x > 45.0 )
    {
        const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
        if ( opp_goalie
             && opp_goalie->distFromSelf() < wm.ball().distFromSelf() )
        {
            rcsc::Line2D cross_line( ball_pos, target_point );
            if (  cross_line.dist( opp_goalie->pos() ) < 3.0 )
            {
                agent->debugClient().addMessage( "AvoidGK" );
                agent->debugClient().addLine( ball_pos, target_point );
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    __FILE__": ShootAreaMove. Goalie is on cross line. old target=(%.2f %.2f)",
                                    target_point.x, target_point.y );

                rcsc::Line2D move_line( wm.self().pos(), target_point );
                target_point.x -= 2.0;
            }
        }
    }

    return target_point;
}

double RoleCenterForward::getShootAreaMoveDashPower( rcsc::PlayerAgent * agent,
                                  const rcsc::Vector2D & target_point )
{

    const rcsc::WorldModel & wm = agent->world();

    static bool s_recover = false;

    double dash_power = rcsc::ServerParam::i().maxPower();

    // recover check
    // check X buffer & stamina
    if ( wm.self().stamina() < rcsc::ServerParam::i().effortDecThrValue() + 500.0 )
    {
        if ( wm.self().pos().x > 30.0 )
        {
            s_recover = true;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": recover on" );
        }
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().effortIncThrValue() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": recover off" );
        s_recover = false;
    }

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( opp_min < self_min - 2
         && opp_min < mate_min - 5 )
    {
        dash_power = std::max( rcsc::ServerParam::i().maxPower() * 0.2,
                               wm.self().playerType().staminaIncMax() * 0.8 );
    }
    else if ( wm.ball().pos().x > wm.self().pos().x )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( wm.ball().pos().x > 41.0
              && wm.ball().pos().absY() < rcsc::ServerParam::i().goalHalfWidth() + 5.0 )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.6;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": recovering" );
    }
    else if ( wm.interceptTable()->teammateReachCycle() <= 1
              && wm.ball().pos().x > 33.0
              && wm.ball().pos().absY() < 7.0
              && wm.ball().pos().x < wm.self().pos().x
              && wm.self().pos().x < wm.offsideLineX()
              && wm.self().pos().absY() < 9.0
              && std::fabs( wm.ball().pos().y - wm.self().pos().y ) < 3.5
              && std::fabs( target_point.y - wm.self().pos().y ) > 5.0 )
    {
        dash_power = wm.self().playerType()
            .getDashPowerToKeepSpeed( 0.3, wm.self().effort() );
        dash_power = std::min( rcsc::ServerParam::i().maxPower() * 0.75,
                               dash_power );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": slow for cross. power=%.1f",
                            dash_power );
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.75;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": ball is far" );
    }

    return dash_power;
}

/******************************************************** END FILE *******************************************/
