// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_defensive_half.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_defensive_move.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/body_hold_ball2008.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_intercept2008.h>
#include <rcsc/action/obsolete/body_kick_multi_step.h>

#include "best_shoot_point.h"
#include "bhv_basic_tackle.h"
#include "bhv_danger_area_move.h"
#include "bhv_offensive_move.h"


/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
	if ( Bhv_BasicTackle( 0.85, 75.0 ).execute( agent ) )
	{
	  return;
        }

	const rcsc::PlayerObject * nearest_opp = agent->world().getOpponentNearestToBall( 10 );

	if( agent->world().ball().pos().dist( agent->world().self().pos() ) < 1.0 && nearest_opp && agent->world().ball().pos().dist( nearest_opp->pos() ) < 2.0 )
	{
		doMove( agent );
        }

	const rcsc::PlayerObject * nearest_tmm = agent->world().getTeammateNearestToBall( 10 );
	if ( !agent->world().existKickableTeammate() && !agent->world().existKickableOpponent() &&
              agent->world().ball().pos().x < -10 && nearest_tmm &&
	      nearest_tmm->pos().dist( agent->world().self().pos() ) <= 0.5 )
	{
		Body_Intercept2008().execute( agent );
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
		return;
	}

	if( !agent->world().existKickableTeammate() &&
	     agent->world().self().pos().dist( agent->world().ball().pos() ) <= 10 && nearest_tmm &&
	     nearest_tmm->pos().dist( agent->world().self().pos() ) <= 0.5 )
	{
		Body_Intercept2008().execute( agent );
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
		return;
	}

        doMove( agent );

    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
    const double nearest_opp_dist
        = agent->world().getDistOpponentNearestToSelf( 5 );

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
        if ( nearest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": doKick() danger" );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                Body_AdvanceBall().execute( agent );

                if ( agent->effector().queuedNextBallKickable() )
                {
                    agent->setNeckAction( new rcsc::Neck_ScanField() );
                }
                else
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
                }
            }
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        if ( nearest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": doKick() danger" );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                Body_AdvanceBall().execute( agent );

                if ( agent->effector().queuedNextBallKickable() )
                {
                    agent->setNeckAction( new rcsc::Neck_ScanField() );
                }
                else
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
                }
            }
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
         Bhv_BasicDefensiveKick().execute( agent );
         break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    {
           rcsc::Vector2D posAgent = agent->world().self().pos();
           int      point;
           Vector2D target_point = Best_ShootPoint().Point( agent , point );

          if( point > 80.0 && posAgent.dist( target_point ) < 10.0 )
          {
             Body_KickMultiStep( target_point , rcsc::ServerParam::i().ballSpeedMax() ).execute( agent );
             agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
             break;
          }
    }
    default:
        rcsc::Body_HoldBall2008().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
	if( wm.ball().pos().x < -30 )
	{
		Bhv_DangerAreaMove( home_pos ).execute( agent );
		break;
	}
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
//         Bhv_DefensiveMove( home_pos ).execute( agent );
//         break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
// 	Bhv_OffensiveMove( home_pos ).execute( agent );
// 	break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}
