// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA
 Modified Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_forward.h"

#include "strategy.h"


#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"


#include <iostream>
#include <math.h>

#include <rcsc/geom/sector_2d.h>

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#include <rcsc/action/obsolete/body_kick_multi_step.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/bhv_shoot2008.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_hold_ball2008.h>
#include <rcsc/action/body_smart_kick.h>
#include "body_clear_ball2008.h"
#include <rcsc/action/body_intercept2008.h>


#include "bhv_penalty_kick.h"
#include "WorldModel.h"
#include "bhv_dribble.h"
#include "bhv_dribble2009.h"
#include "body_penalty_area_pass.h"
#include "best_shoot_point.h"
#include "body_kick_to_corner.h"
#include "bhv_cross_to_center.h"
#include "body_go_to_point_2009.h"
#include "bhv_cross.h"
#include "bhv_attacking_move.h"
#include "bhv_self_pass.h"
#include "bhv_shoot2009.h"


bool doCrossAreaAttempt( rcsc::PlayerAgent *  );


/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

	WorldModelHighLevel WM;

    if ( kickable )
    {
	doKick( agent );
    }
    else
    {
	const rcsc::PlayerObject * nearest_tmm = agent->world().getTeammateNearestToBall( 10 );
	if ( !agent->world().existKickableTeammate() &&
	      agent->world().ball().pos().x > 10 && nearest_tmm &&
	      nearest_tmm->pos().dist( agent->world().self().pos() ) <= 0.5 &&
	      agent->world().existTeammateIn( Circle2D( agent->world().ball().pos() , 7 ) , 10 , false ) )
	{
		Body_Intercept2008().execute( agent );
		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
		return;
	}

	doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doKick( rcsc::PlayerAgent * agent )
{
	WorldModelHighLevel WM;
	const rcsc::WorldModel & wm = agent->world();
	const Vector2D posAgent = wm.self().pos();

	const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
	const rcsc::PlayerObject * nearest_opp2= wm.getOpponentNearestToSelf( 9  );
	const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
	const rcsc::PlayerObject * nearest_tmm = wm.getTeammateNearestToSelf( 10 );
	
	rcsc::Vector2D nearest_opp_pos ( -1000.0, 0.0 );
	rcsc::Vector2D nearest_opp_pos2( -1000.0, 0.0 );
	rcsc::Vector2D opp_goalie_pos  ( -1000.0, 0.0 );
	rcsc::Vector2D nearest_tmm_pos ( -1000.0, 0.0 );
	
	
	if ( opp_goalie  ) opp_goalie_pos   = opp_goalie->pos()  ;
	if ( nearest_opp ) nearest_opp_pos  = nearest_opp->pos() ;
	if ( nearest_opp2) nearest_opp_pos2 = nearest_opp2->pos();
	if ( nearest_tmm ) nearest_tmm_pos  = nearest_tmm->pos() ;
	
	
	rcsc::Triangle2D triangle_goal1( wm.self().pos() ,
					rcsc::Vector2D(52.5, 0.0),
					rcsc::Vector2D(52.5, 8.0));
	
	rcsc::Triangle2D triangle_goal2( wm.self().pos() ,
					rcsc::Vector2D(52.5,-8.0),
					rcsc::Vector2D(52.5, 0.0));
	
	rcsc::Triangle2D triangle_goal ( wm.self().pos() ,
					rcsc::Vector2D(52.5,-8.0),
					rcsc::Vector2D(52.5,+8.0));
	
	rcsc::Triangle2D triangle( wm.self().pos() ,
					wm.self().pos() + rcsc::Vector2D(10.0, 4.0),
					wm.self().pos() + rcsc::Vector2D(10.0,-4.0));
	
	Vector2D posMirror( 52.5 , posAgent.y );
	

	switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
	case Strategy::BA_CrossBlock:
	case Strategy::BA_Stopper:
	case Strategy::BA_Danger:
	case Strategy::BA_DribbleBlock:
	case Strategy::BA_DefMidField:
	case Strategy::BA_DribbleAttack:
	{
		Vector2D target_point ( 0.0 , 0.0 );
		if ( Bhv_Cross().get_best_point( agent , &target_point ) )
		{
// 			Triangle2D target_t = WM.Create_Triangle_From_Two_Points( posAgent , target_point , 6 );
			Sector2D  target_s ( posAgent , 0.5 , posAgent.dist( target_point ) / 2.0 ,
					     (target_point - posAgent ).th() - 15 , (target_point - posAgent ).th() + 15 );

			if( wm.countOpponentsIn( target_s , 10 , true ) == 0 )
			{
// 				Bhv_Cross().execute( agent );
				Bhv_Cross().execute( agent );
				break;
			}
			else
			{
				Bhv_Dribble().execute( agent );
				break;
			}
		}
        }
	case Strategy::BA_OffMidField:
		if( posAgent.x > 5.0 )
		{
			doDribbleAttackAreaKick( agent );
			break;
		}
	case Strategy::BA_Cross:
	{
		if( opp_goalie && Circle2D( posAgent , 3.0 ).contains( opp_goalie_pos ) )
		{
			bool is_let_to_cross_ = Bhv_CrossToCenter().get_best_point( agent , NULL );
			if( is_let_to_cross_ )
			{
				Bhv_CrossToCenter().execute( agent );
			}
			else
			{
				Body_KickToCorner( posAgent.y < 0 ).execute( agent );
				agent->setNeckAction( new Neck_TurnToBallOrScan() );
			}
			break;
		}

		if( doCrossAreaAttempt( agent ) )
		{
			break;
		}

		Rect2D rect ( Vector2D( 52.5 , -12.0 ) , Vector2D( 42.0 , 12.0 ) );
		if( wm.countOpponentsIn ( rect , 8 , false ) <= 1 &&
		    wm.countTeammatesIn ( rect , 8 , false ) > 0 &&
		    WM.isBetween( posAgent.absY() , 20 , 34 ) )
		{
			Vector2D cross_to_center_point( 52.5 , 0.0 );
			bool is_let_to_cross = Bhv_CrossToCenter().get_best_point( agent , &cross_to_center_point );
			if( is_let_to_cross )
			{
				AngleDeg cross_to_center_ang = ( cross_to_center_point - posAgent ).th();
				double   dist_to_center      = posAgent.dist( cross_to_center_point );
				Sector2D sector_to_center( posAgent,
								0.5,
								dist_to_center,
								cross_to_center_ang + 20.0,
								cross_to_center_ang - 20.0 );
	
				if( !wm.existOpponentIn( sector_to_center , 10 , false ) &&
				     wm.existTeammateIn( sector_to_center , 10 , false ) &&
				     posAgent.absY() > 15 )
				{
					Bhv_CrossToCenter().execute( agent );
					break;
				}
			}
		}
	}
	case Strategy::BA_ShootChance:
	{
		if( Bhv_Shoot2008().execute( agent ) )
		{
			break;
		}
		if( Bhv_Shoot2009().execute( agent ) )
		{
			break;
		}

		if( WM.isBetween( agent->world().ball().pos().absY() , 0 , 13 ) &&
                    WM.isBetween( agent->world().ball().pos().x , 35.0 , 45.0 ) )
		{
			if( opp_goalie && fabs(posMirror.y) <= 10.0 &&
			    !triangle_goal1.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 , 6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( opp_goalie && fabs(posMirror.y) <= 10.0 &&
				 !triangle_goal2.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 ,-6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}

		if( opp_goalie && Circle2D( posAgent , 7.0 ).contains( opp_goalie_pos ) )
		{
			if( Bhv_Shoot2008().execute( agent ) )
				break;
			else if( opp_goalie &&
				 fabs(posMirror.y) <= 10.0 &&
				 !triangle_goal1.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 , 6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( opp_goalie && fabs(posMirror.y) <= 10.0 && !triangle_goal2.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 ,-6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( !wm.existOpponentIn( Circle2D( posAgent , 5.0 ) , 10 , false ) )
			{
				doDribbleShootArea( agent );
				break;
			}
			else if( !wm.existOpponentIn( triangle_goal , 10 , false ) )
			{
				doDribbleShootArea( agent );
				break;
			}
			else
			{
				Vector2D kick_to_corner_point = Body_KickToCorner( posAgent.y < 0
										 ).kick_to_corner_target();
				AngleDeg kick_to_corner_ang = ( kick_to_corner_point - posAgent ).th();

				double dist_to_corner = posAgent.dist( kick_to_corner_point );

				Sector2D sector_to_corner ( posAgent,
							    0.5,
							    dist_to_corner,
							    kick_to_corner_ang + 20,
							    kick_to_corner_ang - 20 );

				if( !wm.existOpponentIn( sector_to_corner , 10 , false ) &&
				     wm.existTeammateIn( sector_to_corner , 10 , false ) &&
				     posAgent.absY() < 15 )
				{
					Body_KickToCorner( posAgent.y < 0 ).execute( agent );
					break;
				}

				Vector2D cross_to_center_point( 52.5 , 0.0 );
				bool is_let_to_cross = Bhv_CrossToCenter().get_best_point( agent , &cross_to_center_point );
				if( is_let_to_cross )
				{
					AngleDeg cross_to_center_ang = ( cross_to_center_point - posAgent ).th();
					double   dist_to_center      = posAgent.dist( cross_to_center_point );
					Sector2D sector_to_center( posAgent,
								   0.5,
								   dist_to_center,
								   cross_to_center_ang + 20.0,
								   cross_to_center_ang - 20.0 );

					if( !wm.existOpponentIn( sector_to_center , 10 , false ) &&
					     wm.existTeammateIn( sector_to_center , 10 , false ) &&
					     posAgent.absY() > 15 )
					{
						Bhv_CrossToCenter().execute( agent );
						break;
					}
				}

			}
		}

		if( nearest_opp && posAgent.x > nearest_opp_pos.x )
		{
			doDribbleShootArea( agent );
			break;
		}

		Rect2D rect ( Vector2D( 52.5 , -8.0 ) , Vector2D( 44.5 , 8.0 ) );
		if( wm.countOpponentsIn ( rect , 8 , false ) == 0 &&
		  ( ( ( rect.contains( posAgent ) && wm.countTeammatesIn( rect , 8 , false ) > 1 )
		   || (!rect.contains( posAgent ) && wm.countTeammatesIn( rect , 8 , false ) > 0 ) ) &&
		   posAgent.absY() > 20 ) && Bhv_Cross().execute( agent ) )
		{
			break;
		}

		if( WM.isBetween( agent->world().self().pos().x , 20 , 48 ) &&
			 agent->world().self().pos().absY() < 20.0 )
		{
			if( doDribbleShootArea( agent ) )
			{
				break;
			}
		}
	
		if( WM.isBetween( agent->world().ball().pos().absY() , 0 , 18 ) &&
                    WM.isBetween( agent->world().ball().pos().x , 32.0 , 45.0 ) )
		{
			if( opp_goalie && fabs(posMirror.y) <= 10.0 &&
			    !triangle_goal1.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 , 6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( opp_goalie && fabs(posMirror.y) <= 10.0 &&
				 !triangle_goal2.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 ,-6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}
	
		if( agent->world().self().pos().x > 20 )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
	}
	default:
	{

		if( Circle2D(Vector2D(52.5,0.0),18.0).contains(wm.self().pos()) )
		{
			if( Bhv_Shoot2008().execute( agent ) )
			{
				break;
			}
			else if( opp_goalie && fabs(posMirror.y) <= 10.0 && !triangle_goal1.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 , 6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( opp_goalie && fabs(posMirror.y) <= 10.0 && !triangle_goal2.contains( opp_goalie_pos ) )
			{
				rcsc::Body_KickMultiStep( Vector2D( 52.5 ,-6.5 ) , 100 ).execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else
			{
				Vector2D cross_point(0.0,0.0);
				if( agent->world().self().pos().x > 36.5 && Bhv_Cross().get_best_point( agent , &cross_point ) )
				{
					Triangle2D cross_t = WM.Create_Triangle_From_Two_Points( agent->world().self().pos() , cross_point , 4 );
					if( agent->world().countOpponentsIn( cross_t , 10 , false ) == 0 && Bhv_Cross().execute( agent )  )
					{
						;
					}
					else
					{
						Body_PenaltyAreaPass().execute( agent );
					}
				}
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}
		else if( wm.self().stamina() < 2500 )
		{
			if( wm.self().pos().dist(opp_goalie_pos) <= 18.0  )
			{
				if( Bhv_Shoot2009().execute( agent ) )
				{
					break;
				}
				else if( Bhv_Shoot2008().execute( agent ) )
				{
					break;
				}
			}
			else if( WM.getNrInSetInCircle( agent , Circle2D( posAgent , 6 ) ) == 0 )
			{
				Bhv_Dribble2009().execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else
			{
				Vector2D pass_point ( 0.0 , 0.0 );
				if ( Body_Pass().get_best_pass( agent->world() , &pass_point , NULL , NULL ) )
				{
					if( pass_point.x - agent->world().self().pos().x > -2.0 )
					{
						Body_Pass().execute( agent );
						agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
						break;
					}
					else
					{
						Body_ClearBall2008().execute( agent );
						agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
						break;
					}
				}
				else
				{
					
				}
			}
		}
		else if( wm.self().pos().dist(Vector2D(52.5,0.0)) < 18 ||
			 wm.self().pos().dist(opp_goalie_pos) < 7.0 ||
			 wm.self().pos().x > 33.0 )
		{
			if( WM.getNrInSetInTriangle(agent,WM.Create_Triangle_From_Two_Points(wm.self().pos(),
			    Vector2D(52.5,0.0),5),0) == 0 && wm.self().pos().dist(Vector2D(52.5,0.0)) < 16 )
			{
				if( rcsc::Bhv_Shoot2008().execute( agent ) )
				{
					break;
				}
				if( Bhv_Shoot2009().execute( agent ) )
				{
					break;
				}
			}
			else
			{
				if( WM.getNrInSetInCircle  ( agent , Circle2D( posAgent , 4.0 ) ) == 0 )
				{
					if( WM.isBetween( agent->world().self().pos().x , 20 , 48 ) &&
						agent->world().self().pos().absY() < 20.0 )
					{
						if ( doDribbleShootArea( agent ) )
							;
					}
					else
						Bhv_Dribble().execute( agent );

					break;
				}
				else if( fabs(agent->world().self().pos().y) < 17 )
				{
					if ( Body_Pass().execute( agent ) )
						;
					else
						Body_KickToCorner(agent->world().self().pos().y < 0
								 ).execute( agent );
					agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
					break;
				}
				else
				{
					Bhv_CrossToCenter().execute( agent );
					agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
					break;
				}
			}
		}
		else if( wm.self().pos().dist(rcsc::Vector2D(52.5,0.0)) < 20 )
		{
			if( !isOppGoalieInRect( agent , 0 ) )
			{
				if( rcsc::Bhv_Shoot2008().execute( agent ) )
					break;
				if( Bhv_Shoot2009().execute( agent ) )
					break;
			}
			else
			{
				Body_Pass().execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}
		else if( WM.isInTheirPenaltyArea( agent , wm.ball().pos() ) )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else if( nearest_opp && rcsc::Circle2D(wm.self().pos(),4.0).contains(nearest_opp_pos) &&
			wm.self().pos().x < nearest_opp_pos.x && wm.self().pos().x < -10 )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else if( !opp_goalie && wm.self().pos().dist( Vector2D( 52.5 , 0.0 )) <= 30.0 )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else if( wm.self().pos().x > 40.0 )
		{
			if( nearest_opp && wm.self().pos().x > nearest_opp_pos.x  )
			{
				if ( Bhv_Dribble().execute(agent) )
					;
				else
				{
					Bhv_Dribble2009().execute( agent );
				}
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else if( nearest_opp && fabs(wm.self().pos().y) < 13  && 
				rcsc::Circle2D(wm.self().pos(),5).contains(nearest_opp_pos))
			{
				if( nearest_tmm && nearest_tmm_pos.x > wm.self().pos().x - 3 )
				{
					rcsc::Body_KickMultiStep( nearest_tmm_pos , 40 ).execute(agent);
					agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
					break;
				}
				else
				{
					Body_Pass().execute( agent );
					agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
					break;
				}
			}
			else
			{
				if ( Bhv_Dribble().execute(agent) )
					;
				else
				{
					Bhv_Dribble2009().execute( agent );
				}
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}
		else if( nearest_opp &&
			rcsc::Circle2D(wm.ball().pos(),5.0).contains(nearest_opp_pos) &&
			wm.self().pos().x < nearest_opp_pos.x )
		{
			doCrossAreaKick( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else if( nearest_opp && nearest_opp_pos.x < wm.self().pos().x )
		{
			if ( Bhv_Dribble().execute(agent) )
				;
			else
			{
				Bhv_Dribble2009().execute( agent );
			}
			agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
			break;
		}
		else
		{
			if( wm.self().pos().dist(rcsc::Vector2D(52.5,0.0)) > 20.0 )
			{
				if ( Bhv_Dribble().execute(agent) )
					;
				else
				{
					Bhv_Dribble2009().execute( agent );
				}
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
			else
			{
				Body_Pass().execute( agent );
				agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				break;
			}
		}
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	        break;
	}
//        }
//     }
  }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMove( rcsc::PlayerAgent * agent )
{

    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

	rcsc::Vector2D ball_pos
	= wm.ball().pos();

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.5 );
    }

	if( home_pos.dist( agent->world().ball().pos() ) < 4.0 )
		home_pos = ball_pos;

	switch ( Strategy::get_ball_area( base_pos ) ) {
	case Strategy::BA_CrossBlock:
	case Strategy::BA_Stopper:
	case Strategy::BA_Danger:
	case Strategy::BA_DribbleBlock:
	case Strategy::BA_DefMidField:
	   Bhv_AttackingMove( home_pos , true ).execute( agent );
	   break;
	case Strategy::BA_DribbleAttack:
	case Strategy::BA_OffMidField:
	   Bhv_AttackingMove( home_pos , true ).execute( agent );
	   break;
	case Strategy::BA_Cross:
	case Strategy::BA_ShootChance:
	   doShootAreaMove( agent , home_pos );
	   break;
	default:
		Bhv_BasicMove( home_pos ).execute( agent );
		break;
	}
}


bool RoleSideForward::
isOppGoalieInRect( rcsc::PlayerAgent * agent , int iSide )
{
   rcsc::Rect2D rect;
   if( iSide == 1 )
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9,-9.12) , rcsc::Vector2D(52.5,0.0 ) );//right
   else if( iSide == 2 )
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9, 9.12) , rcsc::Vector2D(52.5,0.0 ) );//left
   else
      rect = rcsc::Rect2D( rcsc::Vector2D(46.9,-9.12) , rcsc::Vector2D(52.5,9.12) );//left and right

    const rcsc::PlayerObject * opp_goalie = agent->world().getOpponentGoalie();

    if( !opp_goalie )
       return false;
    else
    {
	if( rect.contains(opp_goalie->pos()))
		return true;
	else
		return false;
     }
}


void
RoleSideForward::doCrossAreaKick( rcsc::PlayerAgent* agent )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doCrossAreaKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel& wm = agent->world();

    //-------------------------------------------------------
    // !!!check stamina!!!
    static bool S_recover_mode = false;
    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.6 )
    {
        S_recover_mode = false;
    }

    //-------------------------------------------------------
    // check opponent fielder & goalie distance
    const double opp_dist
        = ( wm.opponentsFromSelf().empty()
            ? 1000.0
            : wm.opponentsFromSelf().front()->distFromSelf() );

    const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
    bool goalie_near = false;
    bool goalie_verynear = false;
    if ( opp_goalie )
    {
        if ( opp_goalie->pos().x > 32.0
             && opp_goalie->pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth() )
        {
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 4.0 ) )
            {
                goalie_near = true;
            }
            if ( opp_goalie->distFromBall()
                 < ( rcsc::ServerParam::i().catchAreaLength()
                     + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 2.0 ) )
            {
                goalie_verynear = true;
            }
        }
    }

    //-------------------------------------------------------
    if ( goalie_verynear )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: goalie very near. keep away from goalie "
                            ,__FILE__, __LINE__ );
        // kick to the near side corner
        agent->debugClient().addMessage( "XAreaGKNear" );

	Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------

    WorldModelHighLevel WorldModel;
    rcsc::Vector2D pass_point;
    int receiver_unum = rcsc::Unum_Unknown;
    const bool can_pass = rcsc::Body_Pass::get_best_pass(  wm, &pass_point,
                                                              NULL, &receiver_unum );
    const rcsc::PlayerObject * receiver = ( receiver_unum != rcsc::Unum_Unknown
                                            ? WorldModel.getTeammate( receiver_unum , agent )
                                            : NULL );
    bool receiver_free
        = ( can_pass
            && receiver
            &&  ! wm.existOpponentIn( rcsc::Circle2D( receiver->pos(), 5.0 ), 10, true ) );

    if ( can_pass
         && pass_point.x > 37.0
         && receiver_free
         && pass_point.absY() < 12.0
         && ! wm.existOpponentIn( rcsc::Circle2D( pass_point, 6.0 ), 10, true )
         && ( ! opp_goalie
              || opp_goalie->pos().dist( pass_point ) > 6.0 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: CrossAreaKick. very chance pass."
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPassChance1" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    if ( can_pass )
    {
        rcsc::Rect2D shoot_rect = rcsc::Rect2D::from_center( 52.5 - 7.0,
                                                             0.0,
                                                             14.0,
                                                             32.0 );
        if ( shoot_rect.contains( pass_point )
             && receiver_free
             && wm.countOpponentsIn( shoot_rect, 10, false ) <= 2
             && ! wm.existOpponentIn( rcsc::Circle2D( pass_point, 5.0 ), 10, true ) )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: CrossAreaKick. very chance pass(2)."
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "XAreaPassChance2" );
            rcsc::Body_Pass().execute( agent );

            if ( agent->effector().queuedNextBallKickable() )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            return;
        }
    }

    if ( can_pass
         && pass_point.x > 39.0
         && pass_point.absY() < 7.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: can cross pass."
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass1" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    // very side attack
    if ( can_pass
         && wm.self().pos().absY() > 27.0
         && opp_dist < 4.0
         && pass_point.x > 26.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: very side attack.  cross"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass2" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    //-------------------------------------------------------
    // cross to center
    // exist teammate at cross area
    if ( opp_dist < 3.0
         && Bhv_CrossToCenter::get_best_point( agent, NULL ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: teammate in cross area. cross to center"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaCross1" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

#if 1
    //-------------------------------------------------------
    // dribble cource is blocked
    if ( can_pass
         && receiver_free
         && wm.self().pos().x > 40.0
         && wm.self().pos().absY() > 20.0 )
    {
        rcsc::Vector2D rect_center( 47.0,
                                    ( wm.self().pos().absY() + 9.0 ) * 0.5 );
        if ( wm.self().pos().y < 0.0 ) rect_center.y *= -1.0;

        rcsc::Rect2D side_rect
            = rcsc::Rect2D::from_center( rect_center,
                                         10.0,
                                         wm.self().pos().absY() - 9.0 );
        agent->debugClient().addRectangle( side_rect );
        if ( wm.countOpponentsIn( side_rect, 10, false ) >= 3 ) // 2 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: CrossAreaKick. dribble cource blocked."
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "XAreaPassBlocked" );
            rcsc::Body_Pass().execute( agent );

            if ( agent->effector().queuedNextBallKickable() )
            {
                if ( receiver->posCount() > 0 )
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToPlayerOrScan( receiver ) );
                }
                else
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                }
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            return;
        }
    }
#endif

    //-------------------------------------------------------
    if ( S_recover_mode
         && ! goalie_near
         && opp_dist > 4.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaHold" );
        rcsc::Body_HoldBall2008().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }

    //-------------------------------------------------------
    // keep dribble
    if ( ! goalie_near
         && opp_dist > rcsc::ServerParam::i().tackleDist() + 0.5 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: goalie not near and not tackle opp. wing dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaDrib1" );
	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
        return;
    }
    //-------------------------------------------------------
    // forcely pass
    if ( can_pass )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: enforce pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaPass3" );
        rcsc::Body_Pass().execute( agent );

        if ( agent->effector().queuedNextBallKickable() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        return;
    }

    //-------------------------------------------------------
    // not exist opp on my body dir
    rcsc::Sector2D body_sector( wm.self().pos(),
                                0.5, 8.0,
                                wm.self().body() - 25.0,
                                wm.self().body() + 25.0 );
    // oponent check with goalie
    if ( ! wm.existOpponentIn( body_sector, 10, true ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: exsit opp in my body dir. wing dribble"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaDrib2" );
	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
        return;
    }

    //-------------------------------------------------------
    // forcely cross
    if ( S_recover_mode && opp_dist < 2.5 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: enforce cross"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "XAreaCross2" );
        Bhv_CrossToCenter().execute( agent );
        return;
    }

	if ( Bhv_Dribble().execute(agent) )
		;
	else
	{
		Bhv_Dribble2009().execute( agent );
	}
}

/*----------------------------------------------*/
/*!

*/

void
RoleSideForward::doDribbleAttackAreaKick( rcsc::PlayerAgent* agent )
{
    // !!!check stamina!!!
    static bool S_recover_mode = false;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doDribbleAttackAreaKick" );

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 200.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        S_recover_mode = false;
    }

    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 10 );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    rcsc::Vector2D nearest_opp_pos( -1000.0, 0.0 );
    if ( nearest_opp ) nearest_opp_pos = nearest_opp->pos();

#if 0
    for ( int dash_step = 16; dash_step >= 6; dash_step -= 2 )
    {
        if ( doSelfPass( agent, dash_step ) )
        {
            return;
        }
    }
#else
    if ( Bhv_SelfPass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doDribbleAttackAreaKick SelfPass" );
        return;
    }
#endif

    if ( Bhv_Dribble().doLongDribble( agent ) )
    {
        return;
    }

    // recover mode
    if ( S_recover_mode )
    {
        if ( nearest_opp_dist > 5.0 )
        {
            rcsc::Body_HoldBall2008( true, rcsc::Vector2D(36.0, wm.self().pos().y * 0.9 )//0.0)
                                     ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
        else if ( rcsc::Body_Pass().execute( agent ) )
        {
            return;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": dribble attack. exhaust kick to corner" );
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return;
        }
	if ( Bhv_CrossToCenter().execute( agent ) )
	        return;
    }

    if ( nearest_opp_dist < rcsc::ServerParam::i().defaultPlayerSize() * 2.0 + 0.2
         || wm.existKickableOpponent() )
    {
	Vector2D target_pass ( 0.0 , 0.0 );
	if ( Body_Pass().get_best_pass( wm , &target_pass , NULL , NULL ) )
	{
		if( target_pass.x > wm.self().pos().x )
		{
			Body_Pass().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
			return;
		}
		else
		{
			Bhv_Cross().execute( agent );
			agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
			return;
		}
	}

        if ( wm.self().pos().x < 25.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": dribble attack. kick to near side" );
            Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                               ).execute( agent );
        }
        else if ( wm.self().pos().x < 35.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": dribble attack. kick to space for keep away" );
            Body_Pass().execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": dribble attack. opponent very near. cross to center." );
            Bhv_Cross().execute( agent );
            return;
        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return;
    }


    if ( nearest_opp
         && wm.self().pos().x > 20.0
         && nearest_opp->distFromSelf() < 1.0
         && nearest_opp->pos().x > wm.self().pos().x
         && std::fabs( nearest_opp->pos().y - wm.self().pos().y ) < 0.9 )
    {
        Body_KickToCorner( ( wm.self().pos().y < 0.0 )
                           ).execute( agent );

        agent->setNeckAction( new rcsc::Neck_ScanField() );
        agent->debugClient().addMessage( "SF:BaseKickAvoidOpp" );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": dribble attack. kick to near side to avoid opponent" );
        return;
    }

    // wing dribble
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doDribbleAttackAreaKick. dribble attack. normal dribble" );
//     Bhv_Dribble().execute( agent );
	doCrossAreaKick(agent);
}

/*----------------------------------------------------*/
/*!

*/

void
RoleSideForward::doShootAreaMove( rcsc::PlayerAgent* agent,
                                  const rcsc::Vector2D& home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doShootAreaMove" );

    const rcsc::WorldModel & wm = agent->world();

    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    //------------------------------------------------------

    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doShootAreaMove. intercept cycle. self=%d, mate=%d, opp=%d",
                        self_min, mate_min, opp_min );
    if ( ( ! wm.existKickableTeammate() && self_min < 4 )
         || self_min < mate_min
         || ( self_min <= mate_min && self_min > 5 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. get ball" );
        agent->debugClient().addMessage( "SFW:GetBall(1)" );

        rcsc::Body_Intercept2008().execute( agent );
#if 0
        if ( self_min == 4 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Wide() );
        }
        else if ( self_min == 3 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Normal() );
        }

        if ( wm.self().pos().x > 30.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
#else
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
#endif
        return;
    }

    //------------------------------------------------------
    if ( self_min < 20
         && self_min < mate_min
         && self_min < opp_min + 20
         && wm.ball().pos().absY() > 19.0
         //&& wm.interceptTable()->isOurTeamBallPossessor() )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. get ball(2)" );
        agent->debugClient().addMessage( "SFW:GetBall(2)" );

        rcsc::Body_Intercept2008().execute( agent );
#if 0
        if ( self_min == 4 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Wide() );
        }
        else if ( self_min == 3 && opp_min >= 2 )
        {
            agent->setViewAction( new rcsc::View_Normal() );
        }

        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
#else
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
#endif
        return;
    }

    //------------------------------------------------------
    // decide move target point & dash power
    const rcsc::Vector2D target_point = getShootAreaMoveTarget( agent, home_pos );
    const double dash_power = getShootAreaMoveDashPower( agent, target_point );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        __FILE__": doShootAreaMove. go to cross point (%.2f, %.2f)",
                        target_point.x, target_point.y );

    agent->debugClient().addMessage( "ShootMoveP%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );


    if ( wm.self().pos().x > target_point.x + dist_thr*0.5
         && std::fabs( wm.self().pos().x - target_point.x ) < 3.0
         && wm.self().body().abs() < 10.0 )
    {
        agent->debugClient().addMessage( "Back" );
        double back_dash_power
            = wm.self().getSafetyDashPower( -dash_power );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Back Move" );
        agent->doDash( back_dash_power );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power,
                                 5, // cycle
                                 false, // no back dash
                                 true, // save recovery
                                 20.0 // dir thr
                                 ).execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. already target point. turn to front" );
        rcsc::Body_TurnToAngle( 0.0 ).execute( agent );
    }

    if ( wm.self().pos().x > 35.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Neck to Goalie or Scan" );
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": doShootAreaMove. Neck to Ball or Scan" );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

rcsc::Vector2D
RoleSideForward::getShootAreaMoveTarget( rcsc::PlayerAgent * agent,
                                         const rcsc::Vector2D & home_pos )
{

    const rcsc::WorldModel & wm = agent->world();

    rcsc::Vector2D target_point = home_pos;

    int mate_min = wm.interceptTable()->teammateReachCycle();
    rcsc::Vector2D ball_pos = ( mate_min < 10
                                ? wm.ball().inertiaPoint( mate_min )
                                : wm.ball().pos() );
    bool opposite_side = false;
    const PositionType position_type = Strategy::i().getPositionType( wm.self().unum() );

    if ( ( ball_pos.y > 0.0
           && position_type == Position_Left )
         || ( ball_pos.y <= 0.0
              && position_type == Position_Right )
         || ( ball_pos.y * home_pos.y < 0.0
              && position_type == Position_Center )
         )
    {
        opposite_side = true;
    }

    if ( opposite_side
         && ball_pos.x > 40.0 // very chance
         && 6.0 < ball_pos.absY()
         && ball_pos.absY() < 15.0 )
    {
        rcsc::Circle2D goal_area( rcsc::Vector2D( 47.0, 0.0 ),
                                  7.0 );

        if ( ! wm.existTeammateIn( goal_area, 10, true ) )
        {
            agent->debugClient().addMessage( "GoToCenterCross" );

            target_point.x = 47.0;

            if ( wm.self().pos().x > wm.offsideLineX() - 0.5
                 && target_point.x > wm.offsideLineX() - 0.5 )
            {
                target_point.x = wm.offsideLineX() - 0.5;
            }

            if ( ball_pos.absY() < 9.0 )
            {
                target_point.y = 0.0;
            }
            else
            {
                target_point.y = ( ball_pos.y > 0.0
                                   ? 1.0 : -1.0 );
                                   //? 2.4 : -2.4 );
            }

            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. center cross point" );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. exist teammate in cross point" );
        }
    }

    // consider near opponent
    {
        double opp_dist = 200.0;
        const rcsc::PlayerObject * nearest_opp
            = wm.getOpponentNearestTo( target_point, 10, &opp_dist );

        if ( nearest_opp
             && opp_dist < 2.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": ShootAreaMove. Point Blocked. old target=(%.2f, %.2f)",
                                target_point.x, target_point.y );
            agent->debugClient().addMessage( "AvoidOpp" );
            //if ( nearest_opp->pos().x + 3.0 < wm.offsideLineX() - 1.0 )
            if ( nearest_opp->pos().x + 2.0 < wm.offsideLineX() - 1.0  )
            {
                //target_point.x = nearest_opp->pos().x + 3.0;
                target_point.x = nearest_opp->pos().x + 2.0;
            }
            else
            {
                target_point.x = nearest_opp->pos().x - 2.0;
#if 1
                if ( std::fabs( wm.self().pos().y - target_point.y ) < 1.0 )
                {
                    target_point.x = nearest_opp->pos().x - 3.0;
                }
#endif
            }
        }

    }

    // consider goalie
    if ( target_point.x > 45.0 )
    {
        const rcsc::PlayerObject * opp_goalie = wm.getOpponentGoalie();
        if ( opp_goalie
             && opp_goalie->distFromSelf() < wm.ball().distFromSelf() )
        {
            rcsc::Line2D cross_line( ball_pos, target_point );
            if (  cross_line.dist( opp_goalie->pos() ) < 3.0 )
            {
                agent->debugClient().addMessage( "AvoidGK" );
                agent->debugClient().addLine( ball_pos, target_point );
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    __FILE__": ShootAreaMove. Goalie is on cross line. old target=(%.2f %.2f)",
                                    target_point.x, target_point.y );

                rcsc::Line2D move_line( wm.self().pos(), target_point );
                target_point.x -= 2.0;
            }
        }
    }

    return target_point;
}

double RoleSideForward::getShootAreaMoveDashPower( rcsc::PlayerAgent * agent,
                                  const rcsc::Vector2D & target_point )
{

    const rcsc::WorldModel & wm = agent->world();

    static bool s_recover = false;

    double dash_power = rcsc::ServerParam::i().maxPower();

    // recover check
    // check X buffer & stamina
    if ( wm.self().stamina() < rcsc::ServerParam::i().effortDecThrValue() + 500.0 )
    {
        if ( wm.self().pos().x > 30.0 )
        {
            s_recover = true;
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                __FILE__": recover on" );
        }
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().effortIncThrValue() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": recover off" );
        s_recover = false;
    }

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( opp_min < self_min - 2
         && opp_min < mate_min - 5 )
    {
        dash_power = std::max( rcsc::ServerParam::i().maxPower() * 0.2,
                               wm.self().playerType().staminaIncMax() * 0.8 );
    }
    else if ( wm.ball().pos().x > wm.self().pos().x )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( wm.ball().pos().x > 41.0
              && wm.ball().pos().absY() < rcsc::ServerParam::i().goalHalfWidth() + 5.0 )
    {
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.6;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": recovering" );
    }
    else if ( wm.interceptTable()->teammateReachCycle() <= 1
              && wm.ball().pos().x > 33.0
              && wm.ball().pos().absY() < 7.0
              && wm.ball().pos().x < wm.self().pos().x
              && wm.self().pos().x < wm.offsideLineX()
              && wm.self().pos().absY() < 9.0
              && std::fabs( wm.ball().pos().y - wm.self().pos().y ) < 3.5
              && std::fabs( target_point.y - wm.self().pos().y ) > 5.0 )
    {
        dash_power = wm.self().playerType()
            .getDashPowerToKeepSpeed( 0.3, wm.self().effort() );
        dash_power = std::min( rcsc::ServerParam::i().maxPower() * 0.75,
                               dash_power );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": slow for cross. power=%.1f",
                            dash_power );
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxPower() * 0.75;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": ball is far" );
    }

    return dash_power;
}

/********************************************/
/*!
    \brief dribble shootable area.
*/
bool RoleSideForward::doDribbleShootArea( rcsc::PlayerAgent * agent )
{
   const WorldModel& wm  = agent->world ();
   Vector2D posAgent = wm.self().pos();
   Vector2D target_point( 52.5 , 0.0 );
   WorldModelHighLevel WM;

   Rect2D shoot_area( posAgent - Vector2D( 0.0 , 10.0 ) ,
                      Vector2D( 52.5 , posAgent.y + 10.0 ) );

   const PlayerObject * opp_goalie = wm.getOpponentGoalie();

   if( !opp_goalie )
   {
     Body_Dribble2008( target_point,
                       1.0,
                       rcsc::ServerParam::i().maxPower(),
                       10,
                       false
                     ).execute( agent );

     cout<<"TARGET :: "<<target_point<<endl;
     agent->setNeckAction( new rcsc::Neck_ScanField() );
     return true;
   }

   Vector2D opp_goalie_pos = opp_goalie->pos();
   Triangle2D goal_tri_t ( posAgent , Vector2D( 52.5 , 6.0 ) , Vector2D( 52.5 , -6.0 ) );

   if( !shoot_area.contains( opp_goalie_pos ) )	
   {
     if( wm.countOpponentsIn( goal_tri_t , 10 , false ) == 0 )
     {
       target_point = Vector2D( 50.0 , 0.0 );
     }
     else if( wm.countOpponentsIn( shoot_area , 10 , true ) == 0 )
     {
       target_point = shoot_area.center();
     }
     else
     {
       Sector2D sector_f_s ( posAgent , 0.5 , 8.0 , agent->world().self().body() - 20 , agent->world().self().body() + 20 );

       if( wm.countOpponentsIn( sector_f_s , 10 , true ) == 0 )
       {
          target_point = posAgent + Vector2D::polar2vector( 10.0 , agent->world().self().body() );
       }
       else
       {
          const PlayerObject * nearest_opp = wm.getOpponentNearestToSelf(10);
          if( !nearest_opp )
          {
             if( posAgent.x < 38.0 )
             {
                target_point = Vector2D( posAgent.x + 10 , -7.0 * sign( posAgent.y ) );
             }
             else
             {
                if( posAgent.absY() < 14.0 )
                {
                  if( opp_goalie_pos.absY() > 7.0 )
                  {
                    target_point = posAgent + Vector2D( 52.5 , -6.0 * WM.sign ( opp_goalie_pos.y ) );
                  }
                  else
                  {
                    target_point = opp_goalie_pos;
                  }
                }
                else
                {
                  target_point = Vector2D( 50.0 , 0.0 );
                }
             }
          }
          else
          {
            Vector2D nearest_opp_pos = nearest_opp->pos();
            if( posAgent.x > nearest_opp_pos.x )
            {
              target_point = Vector2D( 50.0 , 6.0 * WM.sign( posAgent.y ) );
            }
            else
            {
               if( posAgent.y > nearest_opp_pos.y )
               {
                 if( posAgent.x < 38 )
                 {
                    target_point = posAgent + Vector2D( 10.0 , 7.0 );
                 }
                 else
                 {
                    target_point = posAgent + Vector2D::polar2vector( 10 , -agent->world().self().body() );
                 }
               }
               else
               {

                 if( posAgent.x < 38 )
                 {
                    target_point = posAgent + Vector2D( 10.0 , -7.0 );
                 }
                 else
                 {
                    target_point = posAgent + Vector2D::polar2vector( 10 , -agent->world().self().body() );
                 }
               }
            }
          }
       }
     }
   }
   else
   {

     if( wm.countOpponentsIn( goal_tri_t , 10 , false ) == 0 )
     {
       target_point = Vector2D( 50.0 , 0.0 );
     }
     else if( shoot_area.center().y > opp_goalie_pos.y )
     {
       if( WM.isBetween( posAgent.y , -10  , 10 ) )
       {
         target_point = posAgent + Vector2D( 10 , 10 );
       }
       else
       {
//          if( opp_goalie_pos.x - posAgent.x > 6 )
//          {
           target_point = posAgent + Vector2D( 5.0 ,  16.0 );/*
         }
         else
         {
           target_point = posAgent + Vector2D( -8.0 , 8.0 );
         }*/
       }
     }
     else
     {
       if( WM.isBetween( posAgent.y , -10.0 , 10.0 ) )
       {
         target_point = posAgent + Vector2D( 10.0 , -10.0 );
       }
       else
       {
//          if( opp_goalie_pos.x - posAgent.x > 6 )
//          {
           target_point = posAgent + Vector2D( 5.0 , -16.0 );
//          }
//          else
//          {
//            target_point = posAgent + Vector2D( -8.0 , 8.0 );
//          }
       }
     }

   if( wm.countOpponentsIn( shoot_area , 10 , false ) > 3 )
   {
     return false;
   }

  }

   const rcsc::Rect2D target_rect
       = rcsc::Rect2D::from_center( wm.ball().pos().x + 5.5,
                                    wm.ball().pos().y,
                                    10.0, 12.0 );
   int dash_step = 10;

   const rcsc::Rect2D safety_rect
       = rcsc::Rect2D::from_center( wm.ball().pos().x + 7.0,
                                    wm.ball().pos().y,
                                    14.0, 15.0 );

   if ( wm.existOpponentIn( target_rect, 10, false ) )
   {
       dash_step = 3;
   }


   Body_Dribble2008( target_point,
                     1.0,
                     rcsc::ServerParam::i().maxPower(),
                     dash_step,
                     false
                    ).execute( agent );

   agent->setNeckAction( new rcsc::Neck_ScanField() );
   return true;

}

/*!
  \brief cross area attempts.
  \return return false or true with actions.
*/

bool doCrossAreaAttempt( rcsc::PlayerAgent * agent )
{
    const WorldModel & wm = agent->world();

    WorldModelHighLevel WM;

    Vector2D receiver;

    const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 6 );

    bool pass = Body_Pass().get_best_pass( wm , &receiver , NULL , NULL ) &&
                WM.isInTheirPenaltyArea( agent , receiver ) ;


    // pass in order to avoid opponent
    if ( pass
         && nearest_opp
         && nearest_opp->distFromSelf() < 5.0 )
    {
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        agent->debugClient().addMessage( "CrossAreaAttempt(1)" );
        return true;
    }

    rcsc::Rect2D front_rect( rcsc::Vector2D( wm.self().pos().x,
                                             wm.self().pos().y - 5.0 ),
                             rcsc::Size2D( 10.0, 10.0 ) );
    if ( pass
         && receiver.x > wm.self().pos().x
         && receiver.absY() > 20.0
         && wm.existOpponentIn( front_rect, 10, false ) ) // without goalie
    {
        agent->debugClient().addMessage( "CrossAreaAttempt(2)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if ( pass
         && receiver.x > wm.self().pos().x + 10.0
         && receiver.absY() > 20.0 )
    {
        agent->debugClient().addMessage( "CrossAreaAttempt(3)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    Vector2D cross_point;

    bool cross = Bhv_Cross().get_best_point( agent , & cross_point );

    AngleDeg relative_ang = 0.0;

    rcsc::Sector2D route_sector( Vector2D( 0.0 , 0.0 ) , 0.0 , 0.0, 0.0 , 0.0 );

    double dist_thr = wm.self().pos().dist( cross_point );

    if ( cross )
    {
      relative_ang = ( cross_point - wm.self().pos() ).th();

      route_sector.assign( cross_point,
                           0.5,
                           dist_thr - 1.0,
                           relative_ang - 12.5,
                           relative_ang + 12.5 );
    }

    if ( cross
         && !wm.existOpponentIn( route_sector , 10 , false ) )
    {
        agent->debugClient().addMessage( "CrossToCenter(1)" );
	double end_speed = 1.8;
	double first_speed = 100.0;

	first_speed = rcsc::calc_first_term_geom_series_last
	                   ( end_speed,
		             dist_thr,
		             rcsc::ServerParam::i().ballDecay() );

	#ifdef USE_SMART_KICK
	rcsc::Body_SmartKick( cross_point,
			      first_speed,
			      first_speed * 0.96,
			      3 ).execute( agent );
	#else
	rcsc::Body_KickMultiStep( cross_point,
				  first_speed,
				  false //true // enforce
				).execute( agent );
	#endif

        agent->setNeckAction( new rcsc::Neck_ScanField() );

        return true;
    }

    if ( pass
         && receiver.x > wm.offsideLineX() )
    {
        agent->debugClient().addMessage( "CrossAreaAttempt(4)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if ( Bhv_SelfPass().execute( agent ) )
    {
        agent->debugClient().addMessage( "Self Passing(4)" );
        return true;
    }


    if ( pass )
    {
        agent->debugClient().addMessage( "CrossAreaAttempt(5)" );
        rcsc::Body_Pass().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if ( cross )
    {
        agent->debugClient().addMessage( "CrossToCenter(2)" );
	double end_speed = 1.8;
	double first_speed = 100.0;

	first_speed = rcsc::calc_first_term_geom_series_last
	                   ( end_speed,
		             dist_thr,
		             rcsc::ServerParam::i().ballDecay() );

	#ifdef USE_SMART_KICK
	rcsc::Body_SmartKick( cross_point,
			      first_speed,
			      first_speed * 0.96,
			      3 ).execute( agent );
	#else
	rcsc::Body_KickMultiStep( cross_point,
				  first_speed,
				  false //true // enforce
				).execute( agent );
	#endif
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    if( nearest_opp && nearest_opp->pos().dist( wm.ball().pos() ) <= ServerParam::i().tackleDist() + 0.2 )
    {
      AngleDeg opp_ang = ( nearest_opp->pos() - wm.self().pos() ).th();
      Vector2D drib_target = Vector2D::polar2vector( 10 , -opp_ang );
      rcsc::Sector2D drib_sector( wm.self().pos(),
                                  0.5 ,
                                  10.0,
                                  -opp_ang - 10,
                                  -opp_ang + 10 );

      if( !wm.existOpponentIn( drib_sector , 10 , false ) )
      {
         rcsc::Body_Dribble2008( drib_target,
                                 1.0,
                                 rcsc::ServerParam::i().maxPower(),
                                 10,
                                 false ).execute( agent );
        agent->debugClient().addMessage( "Body_Dribble2008_opposite_of_nearest_opponent(4)" );
         agent->setNeckAction( new rcsc::Neck_ScanField() );
         return true;
      }
    }


    if( Bhv_SelfPass().execute( agent ) )
    {
      agent->debugClient().addMessage( "Bhv_SelfPass(4)" );
      return true;
    }

    AngleDeg relative_ang_goal = ( Vector2D( 49.0 , 0.0 ) - wm.self().pos() ).th();
    rcsc::Sector2D goal_sector( wm.self().pos(),
                                0.5 ,
                                10.0,
                                relative_ang_goal - 10,
                                relative_ang_goal + 10 );

    if ( !wm.existOpponentIn( route_sector , 10 , false ) )
    {
        agent->debugClient().addMessage( "BodyDribble2008(1)" );
        rcsc::Body_Dribble2008( Vector2D( 49.0 , 0.0 ),
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                10,
                                false
                                ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return true;
    }

    return false;
}

