// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA
 Modified Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "sample_player.h"

#include "soccer_role.h"

#include "bhv_goalie_free_kick.h"
#include "bhv_penalty_kick.h"
#include "bhv_pre_process.h"
#include "bhv_set_play.h"
#include "bhv_set_play_kick_in.h"

#include <rcsc/formation/formation.h>
#include <rcsc/action/kick_table.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/freeform_parser.h>

#include <rcsc/common/basic_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/player_param.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/common/say_message_parser.h>
#include <rcsc/common/free_message_parser.h>

#include <rcsc/param/param_map.h>
#include <rcsc/param/cmd_line_parser.h>

#include <boost/shared_ptr.hpp>

#include <sstream>

#include "WorldModel.h"
#include "bhv_mark_close_opponent.h"
#include "move_to_pos.h"

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_turn_to_ball.h>
#include "body_go_to_point_2009.h"
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/bhv_shoot2008.h>
#include <rcsc/action/body_kick_one_step.h>
#include "bhv_shoot2009.h"
#include "bhv_mark_opponent.h"

static int counter = 0;
static int S_last_mark_time = 0;
static int counter_mark = 0 ;
/*-------------------------------------------------------------------*/
/*!

*/
SamplePlayer::SamplePlayer()
    : PlayerAgent()
{
    typedef boost::shared_ptr< rcsc::SayMessageParser > SMP;

    boost::shared_ptr< rcsc::AudioMemory > audio_memory( new rcsc::AudioMemory );

    M_worldmodel.setAudioMemory( audio_memory );

    addSayMessageParser( SMP( new rcsc::BallMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::PassMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::InterceptMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::GoalieMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::OffsideLineMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::DefenseLineMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::WaitRequestMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::PassRequestMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::DribbleMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::BallGoalieMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::OnePlayerMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::BallPlayerMessageParser( audio_memory ) ) );

    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 9 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 8 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 7 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 6 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 5 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 4 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 3 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 2 >( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::FreeMessageParser< 1 >( audio_memory ) ) );

    typedef boost::shared_ptr< rcsc::FreeformParser > FP;

    setFreeformParser( FP( new rcsc::FreeformParser( M_worldmodel ) ) );
}

/*-------------------------------------------------------------------*/
/*!

*/
SamplePlayer::~SamplePlayer()
{

}

/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::initImpl( rcsc::CmdLineParser & cmd_parser )
{
    bool result = rcsc::PlayerAgent::initImpl( cmd_parser );

    rcsc::ParamMap my_params;
#if 0
    std::string formation_conf;
    my_params.add()
        ( &formation_conf, "fconf" )
        ;

    cmd_parser.parse( my_params );
#endif

    if ( cmd_parser.count( "help" ) > 0 )
    {
        my_params.printHelp( std::cout );
        return false;
    }

    if ( ! result )
    {
        return false;
    }

    if ( cmd_parser.failed() )
    {
        std::cerr << "player: ***WARNING*** detected unsupported options: ";
        cmd_parser.print( std::cerr );
        std::cerr << std::endl;
    }

    if ( ! M_strategy.read( config().configDir() ) )
    {
        std::cerr << "***ERROR*** Failed to read team strategy." << std::endl;
        return false;
    }

    //////////////////////////////////////////////////////////////////
    // Add your code here.
    //////////////////////////////////////////////////////////////////



    return true;
}

/*-------------------------------------------------------------------*/
/*!
  main decision
  virtual method in super class
*/
void
SamplePlayer::actionImpl()
{
    if ( audioSensor().trainerMessageTime() == world().time() )
    {
        std::cerr << world().teamName() << ' '
                  << world().self().unum() << ": "
                  << world().time()
                  << " hear trainer message ["
                  << audioSensor().trainerMessage()
                  << "]"
                  << std::endl;
    }

    if ( audioSensor().freeformMessageTime() == world().time() )
    {
//         std::cerr << world().teamName() << ' '
//                   << world().self().unum() << ": "
//                   << world().time()
//                   << " hear freeform message ["
//                   << audioSensor().freeformMessage()
//                   << "]"
//                   << std::endl;
    }

    //////////////////////////////////////////////////////////////
    // check tackle expires
    // check self position accuracy
    // ball search
    // check queued intention
    // check simultaneous kick
    if ( Bhv_PreProcess( M_strategy ).execute( this ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: preprocess done"
                            ,__FILE__, __LINE__ );
        return;
    }

    //////////////////////////////////////////////////////////////
    // create current role
    boost::shared_ptr< SoccerRole >
        role_ptr = M_strategy.createRole( config().playerNumber(),
                                          world() );

    if ( ! role_ptr )
    {
        std::cerr << config().teamName() << ": "
                  << world().self().unum()
                  << " Error. Role is not registerd.\nExit ..."
                  << std::endl;
        M_client->setServerAlive( false );
        return;
    }
    if  ( ! role_ptr->hasFormation() )
    {
        std::cerr << config().teamName() << ": "
                  << world().self().unum()
                  << " Error. does not have formation.\nExit ..."
                  << std::endl;
        M_client->setServerAlive( false );
        return;
    }


    //////////////////////////////////////////////////////////////
    // back pass model


     if ( (world().gameMode().type() == rcsc::GameMode::FreeKick_
        || world().gameMode().type() == rcsc::GameMode::KickIn_ 
        || world().gameMode().type() == rcsc::GameMode::CornerKick_)
        && world().gameMode().side() != world().ourSide()
        &&!world().self().goalie() )
     {
        if( counter_mark > 6 )
	{
		rcsc::Vector2D move_pos
		= M_strategy.getSetPlayPosition( config().playerNumber(),
						role_ptr->formation(), world() );
		if ( world().self().stamina() > 6000 &&
		     Bhv_MarkCloseOpponent( move_pos ).execute( this ) && world().self().unum() > 5 )
		return;

		else if ( world().self().stamina() > 6000 &&
		     Bhv_MarkCloseOpponent( move_pos ).execute( this ) && world().self().unum() < 6 &&
		     world().self().pos().x > -25.0 )
		return;

		else
		Bhv_SetPlay( move_pos ).execute( this );
	}
	else
	{
		setNeckAction( new Neck_ScanField() );
		counter_mark ++;
	}
        return;
    }

     if( (world().gameMode().type() == rcsc::GameMode::BackPass_
       || world().gameMode().type() == rcsc::GameMode::IndFreeKick_ )
       && world().gameMode().side() == world().ourSide() )
     {
		rcsc::Vector2D move_pos
		= M_strategy.getSetPlayPosition( config().playerNumber(),
						role_ptr->formation(),
						world() );
		WorldModelHighLevel WM;
	
		Circle2D circle( world().ball().pos() , 6.0 );
	
		double dist_thr = world().self().pos().dist(world().ball().pos())*0.1;


		if( counter < 50 )
		{
// 			cout<<"-----------------------"<<endl;
			if( world().self().pos().dist(world().ball().pos()) >= 0.5 )
			{
				if( world().self().unum() == 11 )
				{
					if ( !Body_GoToPoint2009(Vector2D( world().ball().pos().x-0.2,world().ball().pos().y),dist_thr,100.0
					   ).execute( this ) )
					        rcsc::Body_TurnToBall().execute( this );
					else
						this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

				}
				else if( world().self().unum() == 10 )
				{
					if ( !Body_GoToPoint2009( Vector2D( world().ball().pos().x-0.3,world().ball().pos().y+4),dist_thr,100.0
					   ).execute( this ))
					        rcsc::Body_TurnToBall().execute( this );
					else
						this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				}
				else if( world().self().unum() == 9  )
				{
					if ( !Body_GoToPoint2009( Vector2D( world().ball().pos().x-0.3,world().ball().pos().y-4),dist_thr,100.0
					   ).execute( this ))
					        rcsc::Body_TurnToBall().execute( this );
					else
						this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

				}
				else
					Bhv_SetPlayKickIn( move_pos ).execute( this );
			}
			static int last_cycle = 0;

			if( fabs(last_cycle - world().time().cycle()) > 5 )
				counter =0;

			if( world().time().cycle() != last_cycle )
			{
				last_cycle = world().time().cycle();
				counter ++;
			}
		}
		else
		{
			if( world().self().unum() > 8 )
			{
				if( world().self().unum() == 11 )
				{
					bool kickable = world().self().isKickable();
					if ( world().existKickableTeammate()
						&& world().teammatesFromBall().front()->distFromBall()
						< world().ball().distFromSelf() )
					{
						kickable = false;
					}

					if( kickable )
						Body_KickOneStep( world().self().pos()+Vector2D( 0.0 , 6.0 ), 1.2 ).execute( this );
					else
						Bhv_SetPlayKickIn( move_pos ).execute( this );

					this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
				}
				else
					this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

			}
			else
				Bhv_SetPlayKickIn( move_pos ).execute( this );
		}

		return;
	}

     if( (world().gameMode().type() == rcsc::GameMode::BackPass_
       || world().gameMode().type() == rcsc::GameMode::IndFreeKick_ )
       && world().gameMode().side() != world().ourSide() )
     {
	Vector2D * move_pos = new Vector2D [11];
	if ( world().ball().pos().absY() < 7 )
	{
		move_pos[0]  = Vector2D( -50.5 , 0.1 );//1
		move_pos[1]  = Vector2D( -50.5 , 6.0 );//2
		move_pos[2]  = Vector2D( -50.5 , 4.0 );//3
		move_pos[3]  = Vector2D( -50.5 , 1.4 );//4
		move_pos[4]  = Vector2D( -50.5 ,-1.4 );//5
		move_pos[5]  = Vector2D( -50.5 ,-3.8 );//6
		move_pos[6]  = Vector2D( -50.5 ,-6.0 );//7
		move_pos[7]  = Vector2D( -48.0 , 5.0 );//8
		move_pos[8]  = Vector2D( -48.0 , 2.5 );//9
		move_pos[9]  = Vector2D( -48.0 ,-2.6 );//10
		move_pos[10] = Vector2D( -48.0 ,-5.0 );//11
	}
	else if( world().ball().pos().y >= 7  )
	{
		move_pos[0]  = Vector2D( -50.5 , 0.1 );//1
		move_pos[1]  = Vector2D( -51.5 , 6.4 );//2
		move_pos[2]  = Vector2D( -51.5 , 5.0 );//3
		move_pos[3]  = Vector2D( -50.5 , 2.4 );//4
		move_pos[4]  = Vector2D( -50.5 ,-1.4 );//5
		move_pos[5]  = Vector2D( -50.5 ,-3.8 );//6
		move_pos[6]  = Vector2D( -50.5 ,-6.0 );//7
		move_pos[7]  = Vector2D( -50.0 , 7.0 );//8
		move_pos[8]  = Vector2D( -50.0 , 4.5 );//9
		move_pos[9]  = Vector2D( -48.0 , 0.0 );//10
		move_pos[10] = Vector2D( -48.0 ,-2.0 );//11
	}
	else
	{
		move_pos[0]  = Vector2D( -50.5 , 0.1 );//1
		move_pos[1]  = Vector2D( -50.5 , 6.0 );//2
		move_pos[2]  = Vector2D( -50.5 , 4.0 );//3
		move_pos[3]  = Vector2D( -50.5 , 1.4 );//4
		move_pos[4]  = Vector2D( -50.5 ,-1.4 );//5
		move_pos[5]  = Vector2D( -51.5 ,-3.8 );//6
		move_pos[6]  = Vector2D( -51.5 ,-6.0 );//7
		move_pos[7]  = Vector2D( -48.0 , 2.0 );//8
		move_pos[8]  = Vector2D( -48.0 , 0.0 );//9
		move_pos[9]  = Vector2D( -48.0 ,-4.5 );//10
		move_pos[10] = Vector2D( -48.0 ,-7.0 );//11
	}

	if( world().self().pos().dist( move_pos[world().self().unum() - 1] ) > 0.5 )
	{
		if ( !Body_GoToPoint( move_pos[world().self().unum()-1],0.5,rcsc::ServerParam::i().maxPower()).execute( this ))
		        rcsc::Body_TurnToBall().execute( this );
		else
			this->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

		delete [] move_pos;
	}
	return;
     }

    //////////////////////////////////////////////////////////////
    // play_on mode
    if ( world().gameMode().type() == rcsc::GameMode::PlayOn  )
    {
// 	enter = true;
	counter = 0;
	counter_mark = 0;
	S_last_mark_time = 0;
	if( world().self().isKickable() && Bhv_Shoot2008().execute( this ) )
		;
	else if( world().self().isKickable() && Bhv_Shoot2009().execute( this ) )
		;
	else
		role_ptr->execute( this );
        return;
    }

    //////////////////////////////////////////////////////////////
    // kick_in or corner_kick
    if ( ( world().gameMode().type() == rcsc::GameMode::KickIn_
           || world().gameMode().type() == rcsc::GameMode::CornerKick_ )
         && world().ourSide() == world().gameMode().side() )
    {
        if ( world().self().goalie() )
        {
            Bhv_GoalieFreeKick().execute( this );
        }
        else
        {
            const rcsc::Vector2D base_pos = world().ball().pos();
            rcsc::Vector2D home_pos
                = role_ptr->formation().getPosition( config().playerNumber(),
                                                     base_pos );
            Bhv_SetPlayKickIn( home_pos ).execute( this );
        }
        return;
    }
/*
    if ( ( world().gameMode().type() == rcsc::GameMode::KickIn_
           || world().gameMode().type() == rcsc::GameMode::CornerKick_
           || world().gameMode().type() == rcsc::GameMode::FreeKick_)
         && world().ourSide() != world().gameMode().side() )
    {
        if ( world().self().goalie() )
        {
            const rcsc::Vector2D base_pos = world().ball().pos();
            rcsc::Vector2D home_pos
                = role_ptr->formation().getPosition( config().playerNumber(),
                                                     base_pos );
            Bhv_SetPlay( home_pos ).execute( this );
        }
        else
        {
            if ( fabs( S_last_mark_time - world().time().cycle() ) > 200 )
            {
              S_last_mark_time = world().time().cycle();
            }

            const rcsc::Vector2D base_pos = world().ball().pos();
            rcsc::Vector2D home_pos
                = role_ptr->formation().getPosition( config().playerNumber(),
                                                     base_pos );
            if( world().self().stamina() > 5500 )
               Bhv_MarkOpponent( home_pos , S_last_mark_time ).execute( this );
            else
               Bhv_SetPlay( home_pos ).execute( this );
        }
        return;
    }

*/

    //////////////////////////////////////////////////////////////
    // penalty kick mode
    if ( world().gameMode().isPenaltyKickMode() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: penalty kick"
                            ,__FILE__, __LINE__ );
        Bhv_PenaltyKick().execute( this );
        return;
    }

    //////////////////////////////////////////////////////////////
    // goalie free kick mode
    if ( world().self().goalie() )
    {
        Bhv_GoalieFreeKick().execute( this );
        return;
    }

    //////////////////////////////////////////////////////////////
    // other set play mode

    rcsc::Vector2D move_pos
        = M_strategy.getSetPlayPosition( config().playerNumber(),
                                         role_ptr->formation(),
                                         world() );

    //////////////////////////////////////////////////////////////

    Bhv_SetPlay( move_pos ).execute( this );
}

/*-------------------------------------------------------------------*/
/*!
  communication decision.
  virtual method in super class
*/
void
SamplePlayer::communicationImpl()
{
    if ( ! config().useCommunication()
         || world().gameMode().type() == rcsc::GameMode::BeforeKickOff
         || world().gameMode().type() == rcsc::GameMode::AfterGoal_
         || world().gameMode().isPenaltyKickMode() )
    {
        return;
    }

    sayBall();
    sayDefenseLine();
    sayGoalie();
    sayIntercept();
    sayOffsideLine();
    saySelf();

#if 0
    const int len = effector().getSayMessageLength();
    switch ( len ) {
    case 7:
        addSayMessage( new rcsc::FreeMessage< 2 >( "10" ) );
        break;
    case 6:
        addSayMessage( new rcsc::FreeMessage< 3 >( "210" ) );
        break;
    case 5:
        addSayMessage( new rcsc::FreeMessage< 4 >( "3210" ) );
        break;
    case 4:
        addSayMessage( new rcsc::FreeMessage< 5 >( "43210" ) );
        break;
    case 3:
        addSayMessage( new rcsc::FreeMessage< 6 >( "543210" ) );
        break;
    case 2:
        addSayMessage( new rcsc::FreeMessage< 7 >( "6543210" ) );
        break;
    case 1:
        addSayMessage( new rcsc::FreeMessage< 8 >( "76543210" ) );
        break;
    default:
        break;
    }
#endif

    attentiontoSomeone();
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::sayBall()
{
    // ball Info: seen at current
    const int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::BallMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize()
         || world().ball().posCount() > 0
         || world().ball().velCount() > 0 )
    {
        return false;
    }

    const rcsc::PlayerObject * ball_nearest_teammate = NULL;
    const rcsc::PlayerObject * second_ball_nearest_teammate = NULL;

    const rcsc::PlayerPtrCont::const_iterator t_end = world().teammatesFromBall().end();
    for ( rcsc::PlayerPtrCont::const_iterator t = world().teammatesFromBall().begin();
          t != t_end;
          ++t )
    {
        if ( (*t)->isGhost() || (*t)->posCount() >= 10 ) continue;

        if ( ! ball_nearest_teammate )
        {
            ball_nearest_teammate = *t;
        }
        else if ( ! second_ball_nearest_teammate )
        {
            second_ball_nearest_teammate = *t;
            break;
        }
    }

    bool send_ball = false;
    if ( world().self().isKickable()  // I am kickable
         || ( ! ball_nearest_teammate
              || ( ball_nearest_teammate->distFromBall()
                   > world().ball().distFromSelf() - 3.0 ) ) // I am nearest to ball
         || ( second_ball_nearest_teammate
              && ( ball_nearest_teammate->distFromBall()  // nearest to ball teammate is
                   > rcsc::ServerParam::i().visibleDistance() - 0.5 ) // over vis dist
              && ( second_ball_nearest_teammate->distFromBall()
                   > world().ball().distFromSelf() - 5.0 ) ) ) // I am second
    {
        send_ball = true;
    }

    // send ball & goalie
    if ( send_ball
         && static_cast< int >( len + rcsc::BallGoalieMessage::slength() )
         <= rcsc::ServerParam::i().playerSayMsgSize()
         && world().ball().pos().x > 34.0
         && world().ball().pos().absY() < 20.0 )
    {
        const rcsc::PlayerObject * opp_goalie = world().getOpponentGoalie();
        if ( opp_goalie
             && opp_goalie->posCount() == 0
             && opp_goalie->bodyCount() == 0
             && opp_goalie->unum() != rcsc::Unum_Unknown
             && opp_goalie->distFromSelf() < 25.0
             && opp_goalie->pos().x > 52.5 - 16.0
             && opp_goalie->pos().x < 52.5
             && opp_goalie->pos().absY() < 20.0 )
        {
            addSayMessage( new rcsc::BallGoalieMessage( effector().queuedNextBallPos(),
                                                        effector().queuedNextBallVel(),
                                                        opp_goalie->pos(),
                                                        opp_goalie->body() ) );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": say ball/goalie status" );
            debugClient().addMessage( "SayG" );
            return true;
        }
    }

    // send only ball
    if ( send_ball )
    {
        addSayMessage( new rcsc::BallMessage( effector().queuedNextBallPos(),
                                              effector().queuedNextBallVel() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say ball status" );
        debugClient().addMessage( "Sayb" );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::sayGoalie()
{
    // goalie info: ball is in chance area
    int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::GoalieMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize()
         || world().ball().pos().x < 34.0
         || world().ball().pos().absY() > 20.0 )
    {
        return false;
    }

    const rcsc::PlayerObject * opp_goalie = world().getOpponentGoalie();
    if ( opp_goalie
         && opp_goalie->posCount() == 0
         && opp_goalie->bodyCount() == 0
         && opp_goalie->unum() != rcsc::Unum_Unknown
         && opp_goalie->distFromSelf() < 25.0 )
    {
        addSayMessage( new rcsc::GoalieMessage( opp_goalie->unum(),
                                                opp_goalie->pos(),
                                                opp_goalie->body() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say goalie info: %d pos=(%.1f %.1f) body=%.1f",
                            opp_goalie->unum(),
                            opp_goalie->pos().x,
                            opp_goalie->pos().y,
                            opp_goalie->body().degree() );
        debugClient().addMessage( "Sayg" );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::sayIntercept()
{
    // intercept info
    int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::InterceptMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize()
         || world().ball().posCount() > 0
         || world().ball().velCount() > 0 )
    {
        return false;
    }

    int self_min = world().interceptTable()->selfReachCycle();
    int mate_min = world().interceptTable()->teammateReachCycle();
    int opp_min = world().interceptTable()->opponentReachCycle();

    if ( world().self().isKickable() )
    {
        double next_dist =  effector().queuedNextMyPos().dist( effector().queuedNextBallPos() );
        if ( next_dist > world().self().kickableArea() )
        {
            self_min = 10000;
        }
        else
        {
            self_min = 0;
        }
    }

    if ( self_min <= mate_min
         && self_min <= opp_min
         && self_min <= 10 )
    {
        addSayMessage( new rcsc::InterceptMessage( true,
                                                   world().self().unum(),
                                                   self_min ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say self intercept info %d",
                            self_min );
        debugClient().addMessage( "Sayi_self" );
        return true;
    }
#if 0
    else
    {
        {
            const rcsc::PlayerObject * mate = world().interceptTable()->fastestTeammate();

            if ( mate
                 && mate->unum() != rcsc::Unum_Unknown
                 && mate->posCount() <= 0
                 && mate->distFromSelf() <= 15.0
                 && mate_min <= 10
                 && mate_min <= self_min + 1
                 && mate_min <= opp_min + 1 )
            {
                addSayMessage( new rcsc::InterceptMessage( true,
                                                           mate->unum(),
                                                           mate_min ) );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    __FILE__": say teammate intercept info %d",
                                    mate_min );
                debugClient().addMessage( "Sayi_our" );
                return true;
            }
        }
        {
            const rcsc::PlayerObject * opp = world().interceptTable()->fastestOpponent();
            if ( opp
                 && opp->unum() != rcsc::Unum_Unknown
                 && opp->posCount() <= 0
                 && opp->distFromSelf() <= 15.0
                 && opp_min <= 10
                 && opp_min <= self_min + 1
                 && opp_min <= mate_min + 1 )
            {
                addSayMessage( new rcsc::InterceptMessage( false,
                                                           opp->unum(),
                                                           opp_min ) );
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    __FILE__": say opponent intercept info %d",
                                    opp_min );
                debugClient().addMessage( "Sayi_opp" );
                return true;
            }
        }
    }
#endif
    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::sayOffsideLine()
{
    int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::OffsideLineMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize() )
    {
        return false;
    }

    if ( world().offsideLineCount() <= 1
         && world().opponentsFromSelf().size() >= 8
         && 0.0 < world().ball().pos().x
         && world().ball().pos().x < 37.0
         && world().ball().pos().x > world().offsideLineX() - 20.0
         && std::fabs( world().self().pos().x - world().offsideLineX() ) < 20.0
         )
    {
        addSayMessage( new rcsc::OffsideLineMessage( world().offsideLineX() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say offside line %.1f",
                            world().offsideLineX() );
        debugClient().addMessage( "Sayo" );
        return true;
    }

    return false;
}


/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::sayDefenseLine()
{
    int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::DefenseLineMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize()
         || std::fabs( world().self().pos().x - world().defenseLineX() ) > 40.0
         )
    {
        return false;
    }

    int opp_min = world().interceptTable()->opponentReachCycle();

    rcsc::Vector2D opp_trap_pos = world().ball().inertiaPoint( opp_min );

    if ( world().self().goalie()
         && -40.0 < opp_trap_pos.x
         && opp_trap_pos.x > 10.0 )
    {
        addSayMessage( new rcsc::DefenseLineMessage( world().defenseLineX() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say defense line %.1f",
                            world().defenseLineX() );
        debugClient().addMessage( "Sayd" );
        return true;
    }

    return false;
}


/*-------------------------------------------------------------------*/
/*!

*/
bool
SamplePlayer::saySelf()
{
    const int len = effector().getSayMessageLength();
    if ( static_cast< int >( len + rcsc::DefenseLineMessage::slength() )
         > rcsc::ServerParam::i().playerSayMsgSize()
         || std::fabs( world().self().pos().x - world().defenseLineX() ) > 40.0
         )
    {
        return false;
    }

    bool send_self = false;

    if ( len > 0 )
    {
        send_self = true;
    }

    if ( ! send_self )
    {
        const rcsc::PlayerObject * ball_nearest_teammate = NULL;
        const rcsc::PlayerObject * second_ball_nearest_teammate = NULL;

        const rcsc::PlayerPtrCont::const_iterator t_end = world().teammatesFromBall().end();
        for ( rcsc::PlayerPtrCont::const_iterator t = world().teammatesFromBall().begin();
              t != t_end;
              ++t )
        {
            if ( (*t)->isGhost() || (*t)->posCount() >= 10 ) continue;

            if ( ! ball_nearest_teammate )
            {
                ball_nearest_teammate = *t;
            }
            else if ( ! second_ball_nearest_teammate )
            {
                second_ball_nearest_teammate = *t;
                break;
            }
        }

        if ( ball_nearest_teammate
             && ball_nearest_teammate->distFromBall() < world().ball().distFromSelf()
             && ( ! second_ball_nearest_teammate
                  || second_ball_nearest_teammate->distFromBall() > world().ball().distFromSelf() )
             )
        {
            send_self = true;
        }
    }

    if ( send_self )
    {
        addSayMessage( new rcsc::OnePlayerMessage( world().self().unum(),
                                                   effector().queuedNextMyPos(),
                                                   effector().queuedNextMyBody() ) );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": say self status" );
        debugClient().addMessage( "Say1_self" );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
SamplePlayer::attentiontoSomeone()
{
    if ( world().self().pos().x > world().offsideLineX() - 15.0
         && world().interceptTable()->selfReachCycle() <= 3 )
    {
        std::vector< const rcsc::PlayerObject * > candidates;
        const rcsc::PlayerPtrCont::const_iterator end = world().teammatesFromSelf().end();
        for ( rcsc::PlayerPtrCont::const_iterator p = world().teammatesFromSelf().begin();
              p != end;
              ++p )
        {
            if ( (*p)->goalie() ) continue;
            if ( (*p)->unum() == rcsc::Unum_Unknown ) continue;
            if ( (*p)->pos().x > world().offsideLineX() + 1.0 ) continue;

            if ( (*p)->distFromSelf() > 20.0 ) break;

            candidates.push_back( *p );
        }

        const rcsc::PlayerObject * target_teammate = NULL;
        double max_x = -100.0;
        for ( std::vector< const rcsc::PlayerObject * >::const_iterator p = candidates.begin();
              p != candidates.end();
              ++p )
        {
            if ( (*p)->pos().x > max_x )
            {
                max_x = (*p)->pos().x;
                target_teammate = *p;
            }
        }

        if ( target_teammate )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": attentionto most front teammate",
                                world().offsideLineX() );
            debugClient().addMessage( "AttFrontMate" );
            doAttentionto( world().ourSide(), target_teammate->unum() );
            return;
        }

        // maybe ball owner
        if ( world().self().attentiontoUnum() > 0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": attentionto off. maybe ball owner",
                                world().offsideLineX() );
            debugClient().addMessage( "AttOffBOwner" );
            doAttentiontoOff();
        }
        return;
    }

    const rcsc::PlayerObject * fastest_teammate
        = world().interceptTable()->fastestTeammate();

    int self_min = world().interceptTable()->selfReachCycle();
    int mate_min = world().interceptTable()->teammateReachCycle();
    int opp_min = world().interceptTable()->opponentReachCycle();

    if ( ! fastest_teammate )
    {
        if ( world().self().attentiontoUnum() > 0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": attentionto off. no fastest teammate",
                                world().offsideLineX() );
            debugClient().addMessage( "AttOffNoMate" );
            doAttentiontoOff();
        }
        return;
    }

    if ( mate_min < self_min
         && mate_min <= opp_min
         && fastest_teammate->unum() != rcsc::Unum_Unknown )
    {
        // set attention to ball nearest teammate
        debugClient().addMessage( "AttBallOwner" );
        doAttentionto( world().ourSide(), fastest_teammate->unum() );
        return;
    }

    debugClient().addMessage( "AttOff" );
    doAttentiontoOff();
}


/*-------------------------------------------------------------------*/
/*!

*/
void
SamplePlayer::handleServerParam()
{
    if ( rcsc::KickTable::instance().createTables() )
    {
//         std::cerr << world().teamName() << ' '
//                   << world().self().unum() << ": "
//                   << " KickTable crated."
//                   << std::endl;
    }
    else
    {
        std::cerr << world().teamName() << ' '
                  << world().self().unum() << ": "
                  << " KickTable failed..."
                  << std::endl;
        M_client->setServerAlive( false );
    }
}
